unit HrastUnit;

// version: 20
// Copyright  2014 by HrastProgrammer. All rights reserved.
// This unit contains WREN synthesizer modules developed by HrastProgrammer.
// Can be distributed only as part of the complete WREN source package.
//
// version  8: Blue Hell : changed module registration procedure.
// version  9: Blue Hell : layout changes.
// version 10: Blue Hell : some code changes.
// version 11: Blue Hell : added constructors for the Fast/Slow selection.
// version 12: Blue Hell : removal of outcommented code and object renaming.
// version 13: Blue Hell : added un-attenuated FM to HrastSVF.
// version 14: Blue Hell : changed the lights gatherer.
// version 15: Blue Hell : added some normalization to the filter.
// version 16: Blue Hell : renamed some stuff to make things consistent with other modules - theis will break some patches
// version 17: Blue Hell : added unattenuated freq input to osc, added morph and shape modulation to osc.
// version 18: Blue Hell : modified for the new way to couple pin names to memory locations
// version 19: Blue Hell : added am input on HrastFilter and am on HrastOsc output.
// version 20: Blue Hell : modified for splitting out inputs into params and inputs

interface

uses

  KnobsUtils, KnobsConversions, Module_Defs, Module_Implementations;

type

  TModHrastOsc = class( TMod)
  strict private
  const
    p_frequency   =  0;
    p_morph       =  1;
    p_shape       =  2;
    p_bandlimit   =  3;
    p_pw          =  4;
    p_pwmlevel    =  5;
    p_fmlevel     =  6;
    p_pmlevel     =  7;
    p_vintage     =  8;
    p_level       =  9;
    p_cents       = 10;
    p_shapemodamt = 11;
    p_morphmodamt = 12;
  const
    i_pwm         =  0;
    i_fm          =  1;
    i_phase       =  2;
    i_sync        =  3;
    i_shapemod    =  4;
    i_morphmod    =  5;
    i_freq        =  6;
    i_am          =  7;
  const
    o_out         =  0;
  private
    FZ        : TSignal;
    FPhase    : TSignal;
    FPrevSync : Boolean;
  public
    procedure   CreateIO;                                                                                      override;
    procedure   SetDefaults;                                                                                   override;
    procedure   DoTick;                                                                                        override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);                         override;
  end;


  TModHrastSVF = class( TMod)
  strict private
  const
    p_frequency  =  0;
    p_q          =  1;
    p_fmlevel    =  2;
    p_qmlevel    =  3;
    p_inlevel    =  4;
    p_distortion =  5;
    p_lowpass    =  6;
    p_bandpass   =  7;
    p_highpass   =  8;
    p_level12    =  9;
    p_level24    = 10;
  const
    i_in         =  0;
    i_fm         =  1;
    i_qm         =  2;
    i_freq       =  3;
    i_am         =  4;
  const
    o_lp         =  0;
    o_bp         =  1;
    o_hp         =  2;
    o_br         =  3;
    o_12         =  4;
    o_24         =  5;
  private
    FBand12 : TSignal;
    FLow12  : TSignal;
    FBand24 : TSignal;
    FLow24  : TSignal;
  public
    procedure   CreateIO;                                                                                      override;
    procedure   SetDefaults;                                                                                   override;
    procedure   DoTick;                                                                                        override;
  end;



  procedure CreateTables; // BLue Hell : renamed initialize to CreateTables and made it public; table re-init is needed
                          //             after changing the system rate



implementation



uses

  Globals, Math, SysUtils;

const

  TableSize = 8192;

var

  Pow2Table : array [ 0 .. TableSize + 1] of TSignal;
  SineTable : array [ 0 .. TableSize + 1] of TSignal;


  function Pow2Freq( Freq: TSignal): TSignal;
  var
    n: Integer;
    f: TSignal;
  begin
    f := Abs( Freq) * TableSize;

    if f > TableSize - 1
    then f := TableSize - 1;

    n      := Trunc( f);
    f      := f - n;
    Result := Pow2Table[ n];
    Result := Result + f * ( Pow2Table[ n + 1] - Result);
  end;


  function Sine( Freq: TSignal): TSignal;
  var
    n: Integer;
    f: TSignal;
  begin
    f := Abs( Freq) * TableSize;

    if f > TableSize - 1
    then f := TableSize - 1;

    n      := Trunc( f);
    f      := f - n;
    Result := SineTable[ n];
    Result := Result + f * ( SineTable[ n + 1] - Result);
  end;


{ ========
  TModHrastOsc = class( TMod)
  private
    FZ        : TSignal;
    FPhase    : TSignal;
    FPrevSync : Boolean;
  public
}

    procedure   TModHrastOsc.CreateIO; // override;
    begin
      FIsFast := True;
      FIsSlow := False;

      AddParam ( p_frequency   , 'frequency'  );
      AddParam ( p_morph       , 'morph'      );
      AddParam ( p_shape       , 'shape'      );
      AddParam ( p_bandlimit   , 'bandlimit'  );
      AddParam ( p_pw          , 'pw'         );
      AddParam ( p_pwmlevel    , 'pwmlevel'   );
      AddParam ( p_fmlevel     , 'fmlevel'    );
      AddParam ( p_pmlevel     , 'pmlevel'    );
      AddParam ( p_vintage     , 'vintage'    );
      AddParam ( p_level       , 'level'      );
      AddParam ( p_cents       , 'cents'      );
      AddParam ( p_shapemodamt , 'shapemodamt');
      AddParam ( p_morphmodamt , 'morphmodamt');

      AddInput ( i_pwm         , 'pwm'        );
      AddInput ( i_fm          , 'fm'         );
      AddInput ( i_phase       , 'phase'      );
      AddInput ( i_sync        , 'sync'       );
      AddInput ( i_shapemod    , 'shapemod'   );
      AddInput ( i_morphmod    , 'morphmod'   );
      AddInput ( i_freq        , 'freq'       );
      AddInput ( i_am          , 'am'         );

      AddOutput( o_out         , 'out'        );
    end;


    procedure   TModHrastOsc.SetDefaults;
    begin
      FInputs[ i_shapemod] := 1.0;
      FInputs[ i_am      ] := 1.0;
      FDezipperMap := [
        p_frequency  ,
        p_morph      ,
        p_shape      ,
        p_bandlimit  ,
        p_pw         ,
        p_pwmlevel   ,
        p_fmlevel    ,
        p_pmlevel    ,
        p_level      ,
        p_cents      ,
        p_shapemodamt,
        p_morphmodamt
      ];
    end;


    procedure TModHrastOsc.DoTick;
    var
      DeltaPhase: TSignal;
    var
      n   : Integer;
      t   : TSignal;
      y   : TSignal;
      ypw : TSignal;
    begin
      if   not FPrevSync
      and  SignalToLogic( FInputs[i_sync])
      then FPhase := 0;

      FPrevSync  := SignalToLogic( FInputs[ i_sync]);
      DeltaPhase := Pow2Freq( FParams[ p_frequency] + FInputs[ i_freq] + FParams[ p_cents] + FParams[ p_fmlevel] * FInputs[ i_fm]) * System_Rate_Rec;

      t := FPhase + FParams[ p_pmlevel] * FInputs[ i_phase];
      n := Trunc( t);
      t := t - n;

      if t < 0
      then t := t + 1;

      FPhase := FPhase + DeltaPhase;
      n      := Trunc( FPhase);
      FPhase := FPhase - n;

      if   FPhase < 0
      then FPhase := FPhase + 1;

      y := 2 * t - 1 - PolyBlep( t, FParams[ p_bandlimit], DeltaPhase);
      t := t + FParams[ p_pw] + FParams[ p_pwmlevel] * FInputs[ i_pwm];
      n := Trunc( t);
      t := t - n;

      if   t < 0
      then t := t + 1;

      ypw := 2 * t - 1 - PolyBlep( t, FParams[ p_bandlimit], DeltaPhase);
      t   := Clip( FParams[ p_shape] + FInputs[ i_shapemod] * FParams[ p_shapemodamt], 0, 1);

      if   FParams[ p_vintage] <> 0
      then begin                            // Vintage
        if   t > 0.999
        then t := 0.999;

        y   := ( y + 1) * 0.5;
        y   := y * y;
        y   := 2 * y - 1;
        ypw := ( ypw + 1) * 0.5;
        ypw := ypw * ypw;
        ypw := 2 * ypw - 1;
      end;

      FZ := y - ypw * Clip( FParams[ p_morph] + FInputs[ i_morphmod] * FParams[ p_morphmodamt], 0, 1) + t * FZ;

      if   t > 0.99
      then t := 0.99;

      y := FZ * ( t - 1);
      FOutputs[ o_out] := y * FInputs[ i_am] * FParams[ p_level];
    end;


    procedure TModHrastOsc.GatherLights( const aPrefix: string; const aCallback: TLightsHandler);
    begin
      aCallback( Self, MakeInfo( aPrefix, Name, 'syncinput', FInputs[ i_sync]));
    end;


{ ========
  TModHrastSVF = class( TMod)
  private
    FBand12 : TSignal;
    FLow12  : TSignal;
    FBand24 : TSignal;
    FLow24  : TSignal;
  public
}

    procedure   TModHrastSVF.CreateIO; // override;
    begin
      FIsFast := True;
      FIsSlow := False;

      AddParam ( p_frequency , 'frequency' );
      AddParam ( p_q         , 'q'         );
      AddParam ( p_fmlevel   , 'fmlevel'   );
      AddParam ( p_qmlevel   , 'qmlevel'   );
      AddParam ( p_distortion, 'distortion');
      AddParam ( p_inlevel   , 'inlevel'   );
      AddParam ( p_lowpass   , 'lowpass'   );
      AddParam ( p_bandpass  , 'bandpass'  );
      AddParam ( p_highpass  , 'highpass'  );
      AddParam ( p_level12   , 'level12'   );
      AddParam ( p_level24   , 'level24'   );

      AddInput ( i_in        , 'in'        );
      AddInput ( i_fm        , 'fm'        );
      AddInput ( i_qm        , 'qm'        );
      AddInput ( i_freq      , 'freq'      );
      AddInput ( i_am        , 'am'        );

      AddOutput( o_lp        , 'lp'        );
      AddOutput( o_bp        , 'bp'        );
      AddOutput( o_hp        , 'hp'        );
      AddOutput( o_br        , 'br'        );
      AddOutput( o_12        , '12'        );
      AddOutput( o_24        , '24'        );
    end;


    procedure   TModHrastSVF.SetDefaults;
    begin
      FInputs[ i_in  ] := 0.0;
      FInputs[ i_fm  ] := 0.0;
      FInputs[ i_qm  ] := 0.0;
      FInputs[ i_freq] := 0.0;
      FInputs[ i_am  ] := 1.0;
      FDezipperMap:=[
        p_frequency,
        p_q,
        p_fmlevel,
        p_qmlevel,
        p_inlevel,
        p_distortion,
        p_lowpass,
        p_bandpass,
        p_highpass,
        p_level12,
        p_level24
      ]
    end;


    procedure TModHrastSVF.DoTick;
    var
      Oversample : Integer;
      f          : TSignal;
      fb         : TSignal;
      y          : TSignal;
      High12     : TSignal;
      High24     : TSignal;
      Output12   : TSignal;
      Output24   : TSignal;
    begin
      f  := Pow2Freq( FParams[ p_frequency] + FInputs[ i_freq] + FParams[ p_fmlevel] * FInputs[ i_fm]) / System_Rate_2;
      f  := 2 * Sine( f);
      fb := Clip( 1 - ( FParams[ p_q] + FParams[ p_qmlevel] * FInputs[ i_qm]), 0, 1);
      y  := FInputs[ i_am] * FInputs[ i_in] * FParams[ p_inlevel];
      y  := y / ( Abs( y) + ( 1 - Clip( FParams[ p_distortion], 0, 0.999)));

      for Oversample := 1 to 2
      do begin
        High12   := Normalize( y - fb * FBand12 - FLow12);
        FBand12  := Normalize( f * High12  + FBand12);
        FLow12   := Normalize( f * FBand12 + FLow12);
        FLow12   := Normalize( FLow12 * 1.0001 / ( 0.0001 * Abs( FLow12) + 1));
        Output12 := Normalize( FLow12 * FParams[ p_lowpass] + FBand12 * FParams[ p_bandpass] + High12 * FParams[ p_highpass]);

        High24   := Normalize( Output12 - fb * FBand24 - FLow24);
        FBand24  := Normalize( f * High24  + FBand24);
        FLow24   := Normalize( f * FBand24 + FLow24);
        FLow24   := Normalize( FLow24 * 1.0001 / ( 0.01 * Abs( FLow24) + 1));
      end;

      Output24 := FLow24 * FParams[ p_lowpass] + FBand24 * FParams[ p_bandpass] + High24 * FParams[ p_highpass];
      FOutputs[ o_lp] := FLow12;
      FOutputs[ o_bp] := FBand12;
      FOutputs[ o_hp] := High12;
      FOutputs[ o_br] := FLow12 + High12;
      FOutputs[ o_12] := Output12 * FParams[ p_level12];
      FOutputs[ o_24] := Output24 * FParams[ p_level24];
    end;




  procedure CreateTables;
  var
    i: Integer;
    x: TSignal;
  begin
    for i := 0 to TableSize
    do Pow2Table[ i] := UnitsToFrequency( i / TableSize);

    Pow2Table[ TableSize + 1] := Pow2Table[ TableSize];
    x := Pi / TableSize;

    for i := 0 to TableSize - 1
    do SineTable[ i] := Sin( x * i);

    SineTable[ TableSize div 2] := 0;
    SineTable[ TableSize      ] := SineTable[ 0];
    SineTable[ TableSize + 1  ] := SineTable[ 1];
  end;


initialization

  CreateTables;

end.
