unit tube_c;

interface

uses

  System.SysUtils, System.Math,

  KnobsConversions;

//  -------------------------------------------------------------------------------------------------------------------
//
// Original copyright message:
//
//   This file is part of Gnuspeech, an extensible, text-to-speech package, based on real-time, articulatory,
//   speech-synthesis-by-rules.
//
//   Copyright 1991-2012 David R. Hill, Leonard Manzara, Craig Schock
//
//  -------------------------------------------------------------------------------------------------------------------
//
// Translated to Pascal by Blue Hell, copyright note for the changes:
//
//   (C) COPYRIGHT 2019 Blue Hell / Jan Punter
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License version 2 as
//   published by the Free Software Foundation;
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//   For all listed email addresses :
//
//     _dot. to be substituted by a dot      '.'
//     2@t2  to be substituted by an at sign '@'
//
//
//   Blue Hell is a trade mark owned by
//
//     Jan Punter
//     https:// www.bluehell.nl/
//     jan2@t2mail_dot_bluehell_dot_nl
//
//  -------------------------------------------------------------------------------------------------------------------
//
//
//  REVISION INFORMATION  *****************************************************
//  $Author: len $
//  $Date: 1995/04/17 19:51:21 $
//  $Revision: 1.9 $
//  $Source: /cvsroot/softwareTRM/tube.c,v $
//  $State: Exp $
//
//
//  $Log: tube.c,v $
//   * Revision 1.10 2009-04-19 15:14 david
//   * Initial 0.7 release -- added pthread and buffers needed to interface with Objective-C
//   * aComponents created to implement 'Synthesizer' for the Macintosh under OS X
//   * Note that the frication volume final output increases when glottal volume decreases.
//   * This needs to be fixed not
//   *
//   * Revision 1.9 2006/04/01 18:11 david
//   * Remove all things not required for running Synthesizer and split off structures
//   *
//   * Revision 1.8  1995/04/17  19:51:21  len
//   * Temporary fix to frication balance.
//   *
//   * Revision 1.7  1995/03/21  04:52:37  len
//   * Now compiles FAT.  Also adjusted mono and stereo output volume to match
//   * approximately the output volume of the DSP.
//   *
//   * Revision 1.6  1995/03/04  05:55:57  len
//   * Changed controlRate parameter to a Single.
//   *
//   * Revision 1.5  1995/03/02  04:33:04  len
//   * Added amplitude scaling to input of vocal tract and throat, to keep the
//   * software TRM in line with the DSP version.
//   *
//   * Revision 1.4  1994/11/24  05:24:12  len
//   * Added Hi/Low output sample rate case.
//   *
//   * Revision 1.3  1994/10/20  21:20:19  len
//   * Changed nose and mouth aperture filter coefficients, so now specified as
//   * Hz values ( which scale appropriately as the tube length changes), rather
//   * than arbitrary coefficient values ( which don't scale).
//   *
//   * Revision 1.2  1994/08/05  03:12:52  len
//   * Resectioned tube so that it more closely conforms the the DRM proportions.
//   * Also changed frication injection so now allowed from S3 to S10.
//   *
//   * Revision 1.1.1.1  1994/07/07  03:48:52  len
//   * Initial archived version.
//   *
//
//  *****************************************************************************
//
//
//  *****************************************************************************
//  *
//  *     aProgram:       tube
//  *
//  *     Description:   Software ( non-aReal-time) implementation of the Tube
//  *                    Resonance Model for speech production.
//  *
//  *     Author:        Leonard Manzara
//  *
//  *     Date:          July 5th, 1994
//  *
//  *****************************************************************************

type

  TTubeLogger = procedure( const aMsg: string) of object;

  TNoise = class
  private
    FSeed: Double;
  public
    constructor Create;
    function    NextValue: Double;
  end;


  TNoiseFilter = class
  private
    FNoiseX : Double;
  public
    constructor Create;
    function    NextValue( anInput: Double): Double;
  end;


  TMouthNasalFilter = class
  private
    FCoeff       : Double;
    FA10         : Double;
    FA20         : Double;
    FA21         : Double;
    FB11         : Double;
    FB21         : Double;
    FReflectionY : Double;
    FRadiationX  : Double;
    FRadiationY  : Double;
  private
    procedure   SetCoeff( aValue: Double);
  public
    constructor Create;
    function    NextReflectionValue( anInput: Double): Double;
    function    NextRadiationValue ( anInput: Double): Double;
  public
    property    Coeff: Double read FCoeff write SetCoeff;
  end;


  TThroat = class
  private
    FA0         : Double;
    FB1         : Double;
    FY          : Double;
    FSampleRate : Double;
    FCutoff     : Double;
    FGain       : Double;
  private
    procedure   Initialize;
    procedure   SetSampleRate( aValue: Double);
    procedure   SetCutoff    ( aValue: Double);
  public
    constructor Create;
    function    NextValue( anInput: Double): Double;
  public
    property    SampleRate : Double read FSampleRate write SetSampleRate;
    property    Cutoff     : Double read FCutoff     write SetCutoff;
    property    Gain       : Double read FGain       write FGain;
  end;


  TBandpassFilter = class
  private
    FXn1             : Double;
    FXn2             : Double;
    FYn1             : Double;
    FYn2             : Double;
    FAlpha           : Double;
    FBeta            : Double;
    FGamma           : Double;
    FSampleRate      : Double;
    FBandwidth       : Double;
    FCenterFrequency : Double;
  private
    procedure   Initialize;
    procedure   SetSampleRate     ( aValue: Double);
    procedure   SetBandWidth      ( aValue: Double);
    procedure   SetCenterFrequency( aValue: Double);
  public
    constructor Create;
    function    NextValue( anInput: Double): Double;
  public
    property    SampleRate      : Double read FSampleRate      write SetSampleRate;
    property    Bandwidth       : Double read FBandwidth       write SetBandWidth;
    property    CenterFrequency : Double read FCenterFrequency write SetCenterFrequency;
  end;


  TVocalTract = class
  strict private
  const
    TOP                      =  0;
    BOTTOM                   =  1;
    TOTAL_ALPHA_COEFFICIENTS =  3;
    TOTAL_NASAL_SECTIONS     =  6;
    TOTAL_REGIONS            =  8;
    TOTAL_SECTIONS           = 10;
    TOTAL_COEFFICIENTS       = TOTAL_REGIONS;
    TOTAL_NASAL_COEFFICIENTS = TOTAL_NASAL_SECTIONS;

    // OROPHARYNX REGIONS
    R1  = 0;      //  S1
    R2  = 1;      //  S2
    R3  = 2;      //  S3
    R4  = 3;      //  S4 & S5
    R5  = 4;      //  S6 & S7
    R6  = 5;      //  S8
    R7  = 6;      //  S9
    R8  = 7;      //  S10

    // OROPHARYNX SCATTERING JUNCTION COEFFICIENTS (BETWEEN EACH REGION)
    C1  = R1;     //  R1-R2 (S1-S2)
    C2  = R2;     //  R2-R3 (S2-S3)
    C3  = R3;     //  R3-R4 (S3-S4)
    C4  = R4;     //  R4-R5 (S5-S6)
    C5  = R5;     //  R5-R6 (S7-S8)
    C6  = R6;     //  R6-R7 (S8-S9)
    C7  = R7;     //  R7-R8 (S9-S10)
    C8  = R8;     //  R8-AIR (S10-AIR)

    // OROPHARYNX SECTIONS
    S1  = 0;      //  R1
    S2  = 1;      //  R2
    S3  = 2;      //  R3
    S4  = 3;      //  R4
    S5  = 4;      //  R4
    S6  = 5;      //  R5
    S7  = 6;      //  R5
    S8  = 7;      //  R6
    S9  = 8;      //  R7
    S10 = 9;      //  R8

    // FRICATION INJECTION COEFFICIENTS
    FC1                     = 0;      //  S3
    FC2                     = 1;      //  S4
    FC3                     = 2;      //  S5
    FC4                     = 3;      //  S6
    FC5                     = 4;      //  S7
    FC6                     = 5;      //  S8
    FC7                     = 6;      //  S9
    FC8                     = 7;      //  S10
    TOTAL_FRIC_COEFFICIENTS = 8;

    // THREE-WAY JUNCTION ALPHA COEFFICIENTS
    LEFT  = 0;
    RIGHT = 1;
    UPPER = 2;

    // NASAL TRACT SECTIONS
    N1    = 0;
    VELUM = N1;
    N2    = 1;
    N3    = 2;
    N4    = 3;
    N5    = 4;
    N6    = 5;

    // NASAL TRACT COEFFICIENTS
    NC1 = N1;     //  N1-N2
    NC2 = N2;     //  N2-N3
    NC3 = N3;     //  N3-N4
    NC4 = N4;     //  N4-N5
    NC5 = N5;     //  N5-N6
    NC6 = N6;     //  N6-AIR
  private
    FMouthReflectionFilter : TMouthNasalFilter;
    FNasalReflectionFilter : TMouthNasalFilter;
    FSampleRate            : Double;
    FDampingFactor         : Double;
    FFrication             : Double;
    FCurrPtr               : Integer;
    FPrevPtr               : Integer;
    FOropharynx            : array[ 0 .. TOTAL_SECTIONS -1, 0 .. 1, 0 .. 1] of Double;
    FOropharynxCoeff       : array[ 0 .. TOTAL_COEFFICIENTS - 1] of Double;
    FFricationTap          : array[ 0 .. TOTAL_FRIC_COEFFICIENTS - 1] of Double;
    FAlpha                 : array[ 0 .. TOTAL_ALPHA_COEFFICIENTS - 1] of Double;
    FNasal                 : array[ 0 .. TOTAL_NASAL_SECTIONS - 1, 0 .. 1, 0 .. 1] of Double;
    FNasalCoeff            : array[ 0 .. TOTAL_NASAL_COEFFICIENTS - 1  ] of Double;
  private
    procedure   SetSampleRate          ( aValue : Double);
    function    GetMouthReflectionCoeff         : Double;
    procedure   SetMouthReflectionCoeff( aValue : Double);
    function    GetNasalReflectionCoeff         : Double;
    procedure   SetNasalReflectionCoeff( aValue : Double);
  public
    constructor Create;
    destructor  Destroy;                                                                                       override;
    function    NextValue( anInput, aFrication: Double): Double;
  public
    property    SampleRate    : Double read FSampleRate    write SetSampleRate;
    property    DampingFactor : Double read FDampingFactor write FDampingFactor;
    property    Frication     : Double read FFrication     write FFrication;
  public
    property    MouthReflectionCoeff : Double read GetMouthReflectionCoeff write SetMouthReflectionCoeff;
    property    NasalReflectionCoeff : Double read GetNasalReflectionCoeff write SetNasalReflectionCoeff;
  end;


  TCircularBuffer = class
  private
    FData      : array of Double;
    FSize      : Integer;
    FInp       : Integer;
    FOutp      : Integer;
    FFillCount : Integer;
    FFreeCount : Integer;
  public
    constructor Create( aSize: Integer);
    procedure   Push( aValue: Double);
    function    Pop: Double;
    procedure   Clear;
  public
    property    FillCount : Integer read FFillCount;
    property    FreeCount : Integer read FFreeCount;
  end;


  TSynthesizer = class
  private
    FNoise          : TNoise;
    FNoiseFilter    : TNoiseFilter;
    FBandPassFilter : TBandPassFilter;
    FVocalTract     : TVocalTract;
    FThroat         : TThroat;
  private
    FSampleRate     : Double;
    FGlotPitch      : Double;
    FFrequency      : Double;
    FGlotVolume     : Double;
    FAspVolume      : Double;
    FBreathiness    : Double;
    FCrossMix       : Double;
    FModulation     : Boolean;
  private
    procedure   SetSampleRate( aValue : Double);
    procedure   SetGlotPitch ( aValue : Double);
  private
    function    GetBandpassBandwidth                : Double;
    procedure   SetBandpassBandwidth       ( aValue : Double);
    function    GetBandpassCenterFrequency          : Double;
    procedure   SetBandpassCenterFrequency ( aValue : Double);
    function    GetThroatCutoff                     : Double;
    procedure   SetThroatCutoff            ( aValue : Double);
    function    GetThroatGain                       : Double;
    procedure   SetThroatGain              ( aValue : Double);
    function    GetMouthReflectionCoeff             : Double;
    procedure   SetMouthReflectionCoeff    ( aValue : Double);
    function    GetNasalReflectionCoeff             : Double;
    procedure   SetNasalReflectionCoeff    ( aValue : Double);
  private
    procedure   SetDefaults;
  public
    constructor Create;
    destructor  Destroy;                                                                                       override;
    function    NextValue( anInput: Double): Double;
  public
    property    SampleRate  : Double  read FSampleRate  write SetSampleRate;
    property    GlotPitch   : Double  read FGlotPitch   write SetGlotPitch;
    property    GlotVolume  : Double  read FGlotVolume  write FGlotVolume;
    property    AspVolume   : Double  read FAspVolume   write FAspVolume;
    property    Breathiness : Double  read FBreathiness write FBreathiness;
    property    CrossMix    : Double  read FCrossMix    write FCrossMix;
    property    Modulation  : Boolean read FModulation  write FModulation;
  public
    property    BandPassBandwidth       : Double read GetBandpassBandwidth       write SetBandpassBandwidth;
    property    BandpassCenterFrequency : Double read GetBandpassCenterFrequency write SetBandpassCenterFrequency;
    property    ThroatCutoff            : Double read GetThroatCutoff            write SetThroatCutoff;
    property    ThroatGain              : Double read GetThroatGain              write SetThroatGain;
    property    MouthReflectionCoeff    : Double read GetMouthReflectionCoeff    write SetMouthReflectionCoeff;
    property    NasalReflectionCoeff    : Double read GetNasalReflectionCoeff    write SetNasalReflectionCoeff;
  end;



  // Property getter / settr functions

  procedure ConnectLogger( const aLogger: TTubeLogger);

  procedure setGlotPitch   ( value: Double);
  procedure setGlotVol     ( value: Double);
  procedure setAspVol      ( value: Double);
  procedure setFricVol     ( value: Double);
  procedure setfricPos     ( value: Double);
  procedure setFricCF      ( value: Double);
  procedure setFricBW      ( value: Double);
  procedure setRadius      ( value: Double; index: Integer);
  procedure setVelum       ( value: Double);
  procedure setVolume      ( value: Double);
  procedure setWaveformType( value: Integer);
  procedure setTp          ( value: Double);
  procedure setTnMin       ( value: Double);
  procedure setTnMax       ( value: Double);
  procedure setBreathiness ( value: Double);
  procedure setTubeLength  ( value: Double);
  procedure setTemperature ( value: Double);
  procedure setLossFactor  ( value: Double);
  procedure setApScale     ( value: Double);
  procedure setMouthCoef   ( value: Double);
  procedure setNoseCoef    ( value: Double);
  procedure setNoseRadius  ( value: Double; index: Integer);
  procedure setThroatCutoff( value: Double);
  procedure setModulation  ( value: Integer);
  procedure setMixOffset   ( value: Double);

  function  getGlotPitchDefault                   : Double;
  function  getGlotVolDefault                     : Double;
  function  getAspVolDefault                      : Double;
  function  getFricVolDefault                     : Double;
  function  getFricPosDefault                     : Double;
  function  getFricCFDefault                      : Double;
  function  getFricBWDefault                      : Double;
  function  getRadiusDefault( index: Integer)     : Double;
  function  getVelumRadiusDefault                 : Double;
  function  getVolumeDefault                      : Double;
  function  getBalanceDefault                     : Double;
  function  getWaveformDefault                    : Integer;
  function  getTpDefault                          : Double;
  function  getTnMinDefault                       : Double;
  function  getTnMaxDefault                       : Double;
  function  getBreathinessDefault                 : Double;
  function  getTubeLengthDefault                  : Double;
  function  getTemperatureDefault                 : Double;
  function  getLossFactorDefault                  : Double;
  function  getApScaleDefault                     : Double;
  function  getMouthCoefDefault                   : Double;
  function  getNoseCoefDefault                    : Double;
  function  getNoseRadiusDefault( index: Integer) : Double;
  function  getThroatCutoffDefault                : Double;
  function  getThroatVolDefault                   : Double;
  function  getModulationDefault                  : Integer;
  function  getMixOffsetDefault                   : Double;

  function  getGlotPitch                   : Double;
  function  getGlotVol                     : Double;
  function  getAspVol                      : Double;
  function  getFricVol                     : Double;
  function  getFricPos                     : Double;
  function  getFricCF                      : Double;
  function  getFricBW                      : Double;
  function  getRadius( index: Integer)     : Double;
  function  getVelumRadius                 : Double;
  function  getVolume                      : Double;
  function  getBalance                     : Double;
  function  getWaveform                    : Integer;
  function  getTp                          : Double;
  function  getTnMin                       : Double;
  function  getTnMax                       : Double;
  function  getBreathiness                 : Double;
  function  getTubeLength                  : Double;
  function  getTemperature                 : Double;
  function  getLossFactor                  : Double;
  function  getApScale                     : Double;
  function  getMouthCoef                   : Double;
  function  getNoseCoef                    : Double;
  function  getNoseRadius( index: Integer) : Double;
  function  getThroatCutoff                : Double;
  function  getThroatVol                   : Double;
  function  getModulation                  : Integer;
  function  getMixOffset                   : Double;
  function  getActualTubeLength            : Double;
  function  getControlPeriod               : Integer;
  function  getControlRate                 : Double;
  function  getSampleRate                  : Integer;


implementation


  // COMPILE WITH OVERSAMPLING OR PLAIN OSCILLATOR

{$define OVERSAMPLING_OSCILLATOR}


{ ========
  TNoise = class
  private
    FSeed: Double;
  public
}
    constructor TNoise.Create;
    begin
      FSeed := 0.7892347;
    end;


    function    TNoise.NextValue: Double;
    var
      aProduct : Double;
    begin
      aProduct := 377.0 * FSeed;
      FSeed    := aProduct - Floor( aProduct);
      Result   := FSeed - 0.5;
    end;


{ ========
  TNoiseFilter = class
  private
    FNoiseX : Double;
  public
}

    constructor TNoiseFilter.Create;
    begin
      FNoiseX := 0.0;
    end;


    function    TNoiseFilter.NextValue( anInput: Double): Double;
    begin
      Result  := anInput + FNoiseX;
      FNoiseX := anInput;
    end;


{ ========
  TMouthNasalFilter = class
  private
    FCoeff       : Double;
    FA10         : Double;
    FA20         : Double;
    FA21         : Double;
    FB11         : Double;
    FB21         : Double;
    FReflectionY : Double;
    FRadiationX  : Double;
    FRadiationY  : Double;
  public
    property    Coeff: Double read FCoeff write SetCoeff;
  private
}

    procedure   TMouthNasalFilter.SetCoeff( aValue: Double);
    begin
      if aValue <> FCoeff
      then begin
        FCoeff := aValue;
        FB11   := - aValue;
        FA10   :=   1.0 - Abs( FB11);
        FA20   :=   aValue;
        FB21   := - FA20;
        FA21   :=   FB21;
      end;
    end;


//  public

    constructor TMouthNasalFilter.Create;
    begin
      FReflectionY := 0.0;
      FRadiationX  := 0.0;
      FRadiationY  := 0.0;
    end;


    function    TMouthNasalFilter.NextReflectionValue( anInput: Double): Double;
    begin
      Result       := FA10 * anInput + FB11 * FReflectionY;
      FReflectionY := Result;
    end;


    function    TMouthNasalFilter.NextRadiationValue ( anInput: Double): Double;
    begin
      Result      := FA20 * anInput + FA21 * FRadiationX - FB21 * FRadiationY;
      FRadiationX := anInput;
      FRadiationY := Result;
    end;


{ ========
  TThroat = class
  private
    FA0         : Double;
    FB1         : Double;
    FY          : Double;
    FSampleRate : Double;
    FCutoff     : Double;
    FGain       : Double;
  public
    property    SampleRate : Double read FSampleRate write SetSampleRate;
    property    Cutoff     : Double read FCutoff     write SetCutoff;
    property    Gain       : Double read FGain       write FGain;
  private
}

    procedure   TThroat.Initialize;
    begin
      FA0 := 2.0 * FCutoff / FSampleRate;
      FB1 := 1.0 - FA0;
    end;


    procedure   TThroat.SetSampleRate( aValue: Double);
    begin
      if aValue <> FSampleRate
      then begin
        FSampleRate := aValue;
        Initialize;
      end;
    end;


    procedure   TThroat.SetCutoff( aValue: Double);
    begin
      if aValue <> FCutoff
      then begin
        FCutoff := aValue;
        Initialize;
      end;
    end;


//  public

    constructor TThroat.Create;
    begin
      FY := 0.0;
    end;


    function    TThroat.NextValue( anInput: Double): Double;
    begin
      Result := FA0 * anInput + FB1 * FY;
      FY     := Result;
      Result := FGain * Result;
    end;


{ ========
  TBandpassFilter = class
  private
    FXn1             : Double;
    FXn2             : Double;
    FYn1             : Double;
    FYn2             : Double;
    FAlpha           : Double;
    FBeta            : Double;
    FGamma           : Double;
    FSampleRate      : Double;
    FBandwidth       : Double;
    FCenterFrequency : Double;
  public
    property    SampleRate      : Double write SetSampleRate;
    property    BandWidth       : Double write SetBandWidth;
    property    CenterFrequency : Double write SetCenterFrequency;
  private
}

    procedure   TBandpassFilter.Initialize;
    var
      aTan : Double;
      aCos : Double;
    begin
      aTan := Tan( PI * FBandWidth / FSampleRate);
      aCos := Cos( 2.0 * PI * FCenterFrequency / FSampleRate);

      FBeta  := ( 1.0 - aTan) / ( 2.0 * ( 1.0 + aTan));
      FGamma := ( 0.5 + FBeta) * aCos;
      FAlpha := ( 0.5 - FBeta) / 2.0;
    end;


    procedure   TBandpassFilter.SetSampleRate( aValue: Double);
    begin
      if aValue <> FSampleRate
      then begin
        FSampleRate := aValue;
        Initialize;
      end;
    end;


    procedure   TBandpassFilter.SetBandWidth( aValue: Double);
    begin
      if aValue <> FBandWidth
      then begin
        FBandWidth := aValue;
        Initialize;
      end;
    end;


    procedure   TBandpassFilter.SetCenterFrequency( aValue: Double);
    begin
      if aValue <> FCenterFrequency
      then begin
        FCenterFrequency := aValue;
        Initialize;
      end;
    end;


//  public

    constructor TBandpassFilter.Create;
    begin
    end;


    function    TBandpassFilter.NextValue( anInput: Double): Double;
    begin
      Result := 2.0 * (( FAlpha * ( anInput - FXn2)) + ( FGamma * FYn1) - ( FBeta * FYn2));
      FXn2   := FXn1;
      FXn1   := anInput;
      FYn2   := FYn1;
      FYn1   := Result;
    end;


{ ========
  TVocalTract = class
  private
    FSampleRate : Double;
  public
    property    SampleRate : Double read FSampleRate write SetSampleRate;
  private
}

    procedure   TVocalTract.SetSampleRate( aValue: Double);
    begin
      if aValue <> FSampleRate
      then begin
        FSampleRate := aValue;
      end;
    end;


    function    TVocalTract.GetMouthReflectionCoeff: Double;
    begin
      Result := FMouthReflectionFilter.Coeff;
    end;


    procedure   TVocalTract.SetMouthReflectionCoeff( aValue : Double);
    begin
      FMouthReflectionFilter.Coeff := aValue;
    end;


    function    TVocalTract.GetNasalReflectionCoeff: Double;
    begin
      Result := FNasalReflectionFilter.Coeff;
    end;


    procedure   TVocalTract.SetNasalReflectionCoeff( aValue : Double);
    begin
      FNasalReflectionFilter.Coeff := aValue;
    end;


//  public

    constructor TVocalTract.Create;
    begin
      FMouthReflectionFilter := TMouthNasalFilter.Create;
      FNasalReflectionFilter := TMouthNasalFilter.Create;
    end;


    destructor  TVocalTract.Destroy; // override;
    begin
      FreeAndNil( FNasalReflectionFilter);
      FreeAndNil( FMouthReflectionFilter);
    end;


    function    TVocalTract.NextValue( anInput, aFrication: Double): Double;
    var
      i                : Integer;
      j                : Integer;
      k                : Integer;
      Delta            : Double;
      JunctionPressure : Double;
    begin
      FCurrPtr := ( FCurrPtr + 1) mod 2;
      FPrevPtr := ( FPrevPtr + 1) mod 2;

      FOropharynx[ S1][ TOP][ FCurrPtr] := ( FOropharynx[ S1][ BOTTOM][ FPrevPtr] * dampingFactor) + anInput;

      // CALCULATE THE SCATTERING JUNCTIONS FOR S1-S2
      Delta := FOropharynxCoeff[ C1] * ( FOropharynx[ S1][ TOP][ FPrevPtr] - FOropharynx[ S2][ BOTTOM][ FPrevPtr]);
      FOropharynx[ S2][ TOP][ FCurrPtr] := ( FOropharynx[ S1][ TOP][ FPrevPtr] + Delta) * dampingFactor;
      FOropharynx[ S1][ BOTTOM][ FCurrPtr] := ( FOropharynx[ S2][ BOTTOM][ FPrevPtr] + Delta) * dampingFactor;

      // CALCULATE THE SCATTERING JUNCTIONS FOR S2-S3 AND S3-S4

      j := C2;
      k := FC1;

      for i := S2 to S4 - 1
      do begin
        Delta := FOropharynxCoeff[ j] * ( FOropharynx[ i][ TOP][ FPrevPtr] - FOropharynx[ i + 1][ BOTTOM][ FPrevPtr]);
        FOropharynx[ i + 1][ TOP][ FCurrPtr] := (( FOropharynx[ i][ TOP][ FPrevPtr] + Delta) * dampingFactor) + ( FFricationTap[ k] * frication);
        FOropharynx[ i][ BOTTOM][ FCurrPtr] := ( FOropharynx[ i + 1][ BOTTOM][ FPrevPtr] + Delta) * dampingFactor;
        Inc( j);
        Inc( k);
       end;

      // UPDATE 3-WAY JUNCTION BETWEEN THE MIDDLE OF R4 AND NASAL CAVITY
      JunctionPressure := ( FAlpha[ LEFT] * FOropharynx[ S4][ TOP][ FPrevPtr]) + ( FAlpha[ RIGHT] * FOropharynx[ S5][ BOTTOM][ FPrevPtr]) + ( FAlpha[ UPPER] * FNasal[ VELUM][ BOTTOM][ FPrevPtr]);
      FOropharynx[ S4][ BOTTOM][ FCurrPtr] := ( JunctionPressure - FOropharynx[ S4][ TOP][ FPrevPtr]) * dampingFactor;
      FOropharynx[ S5][ TOP][ FCurrPtr] := (( JunctionPressure - FOropharynx[ S5][ BOTTOM][ FPrevPtr]) * dampingFactor) + ( FFricationTap[ FC3] * frication);
      FNasal[ VELUM][ TOP][ FCurrPtr] := ( JunctionPressure - FNasal[ VELUM][ BOTTOM][ FPrevPtr]) * dampingFactor;

      // CALCULATE JUNCTION BETWEEN R4 AND R5 ( S5-S6)
      Delta := FOropharynxCoeff[ C4] * ( FOropharynx[ S5][ TOP][ FPrevPtr] - FOropharynx[ S6][ BOTTOM][ FPrevPtr]);
      FOropharynx[ S6][ TOP][ FCurrPtr] := (( FOropharynx[ S5][ TOP][ FPrevPtr] + Delta) * dampingFactor) + ( FFricationTap[ FC4] * frication);
      FOropharynx[ S5][ BOTTOM][ FCurrPtr] := ( FOropharynx[ S6][ BOTTOM][ FPrevPtr] + Delta) * dampingFactor;

      // CALCULATE JUNCTION INSIDE R5 ( S6-S7) ( PURE DELAY WITH DAMPING)
      FOropharynx[ S7][ TOP][ FCurrPtr] := ( FOropharynx[ S6][ TOP][ FPrevPtr] * dampingFactor) + ( FFricationTap[ FC5] * frication);
      FOropharynx[ S6][ BOTTOM][ FCurrPtr] := FOropharynx[ S7][ BOTTOM][ FPrevPtr] * dampingFactor;

      // CALCULATE LAST 3 INTERNAL JUNCTIONS ( S7-S8, S8-S9, S9-S10)

      j := C5;
      k := FC6;

      for i := S7 to S10 - 1
      do begin
        Delta := FOropharynxCoeff[ j] * ( FOropharynx[ i][ TOP][ FPrevPtr] - FOropharynx[ i + 1][ BOTTOM][ FPrevPtr]);
        FOropharynx[ i + 1][ TOP][ FCurrPtr] := (( FOropharynx[ i][ TOP][ FPrevPtr] + Delta) * dampingFactor) + ( FFricationTap[ k] * frication);
        FOropharynx[ i][ BOTTOM][ FCurrPtr] := ( FOropharynx[ i + 1][ BOTTOM][ FPrevPtr] + Delta) * dampingFactor;
        Inc( j);
        Inc( k);
      end;

      // REFLECTED SIGNAL AT MOUTH GOES THROUGH A LOWPASS FILTER
      FOropharynx[ S10][ BOTTOM][ FCurrPtr] :=  dampingFactor * FMouthReflectionFilter.NextReflectionValue( FOropharynxCoeff[ C8] * FOropharynx[ S10][ TOP][ FPrevPtr]);

      // OUTPUT FROM MOUTH GOES THROUGH A HIGHPASS FILTER
      Result := FMouthReflectionFilter.NextRadiationValue(( 1.0 + FOropharynxCoeff[ C8]) * FOropharynx[ S10][ TOP][ FPrevPtr]);

      // UPDATE NASAL CAVITY
      j := NC1;

      for i := VELUM to N6 - 1
      do begin
        Delta := FNasalCoeff[ j] * ( FNasal[ i][ TOP][ FPrevPtr] - FNasal[ i + 1][ BOTTOM][ FPrevPtr]);
        FNasal[ i + 1][ TOP][ FCurrPtr] := ( FNasal[ i][ TOP][ FPrevPtr] + Delta) * dampingFactor;
        FNasal[ i][ BOTTOM][ FCurrPtr] := ( FNasal[ i + 1][ BOTTOM][ FPrevPtr] + Delta) * dampingFactor;
        Inc( j);
      end;

      // REFLECTED SIGNAL AT NOSE GOES THROUGH A LOWPASS FILTER
      FNasal[ N6][ BOTTOM][ FCurrPtr] := dampingFactor * FNasalReflectionFilter.NextReflectionValue( FNasalCoeff[ NC6] * FNasal[ N6][ TOP][ FPrevPtr]);

      // OUTPUT FROM NOSE GOES THROUGH A HIGHPASS FILTER
      Result := Result + FNasalReflectionFilter.NextRadiationValue(( 1.0 + FNasalCoeff[ NC6]) * FNasal[ N6][ TOP][ FPrevPtr]);
    end;


{ ========
  TCircularBuffer = class
  private
    FData      : array of Double;
    FSize      : Integer;
    FInp       : Integer;
    FOutp      : Integer;
    FFillCount : Integer;
    FFreeCount : Integer;
  public
    property    FillCount : Integer read FFillCount;
    property    FreeCount : Integer read FFreeCount;
  public
}

    constructor TCircularBuffer.Create( aSize: Integer);
    begin
      inherited Create;
      SetLength( FData, aSize);
      FSize := aSize;
      Clear;
    end;


    procedure   TCircularBuffer.Push( aValue: Double);
    begin
      if   FFreeCount > 0
      then begin
        Dec( FFreeCount);
        FData[ FInp] := aValue;
        FInp := ( FInp + 1) mod FSize;
        Inc( FFillCount);
      end;
    end;


    function    TCircularBuffer.Pop: Double;
    begin
      if   FFillCount > 0
      then begin
        Dec( FFillCount);
        Result := FData[ FOutp];
        FOutp  := ( FOutp + 1) mod FSize;
        Inc( FFreeCount);
      end
      else Result := 0.0;
    end;


    procedure   TCircularBuffer.Clear;
    begin
      FFreeCount := 0;
      FFillCount := 0;
      FInp       := 0;
      FOutp      := 0;
      FFreeCount := FSize;
    end;


{ ========
  TSynthesizer = class
  private
    FNoise          : TNoise;
    FNoiseFilter    : TNoiseFilter;
    FBandPassFilter : TBandPassFilter;
    FVocalTract     : TVocalTract;
    FThroat         : TThroat;
  private
    FSampleRate     : Double;
    FGlotPitch      : Double;
    FFrequency      : Double;
    FGlotVolume     : Double;
    FAspVolume      : Double;
    FBreathiness    : Double;
    FCrossMix       : Double;
    FModulation     : Boolean;
    FModulation  : Boolean;
  public
    property    SampleRate  : Double  read FSampleRate  write SetSampleRate;
    property    GlotPitch   : Double  read FGlotPitch   write SetGlotPitch;
    property    GlotVolume  : Double  read FGlotVolume  write FGlotVolume;
    property    AspVolume   : Double  read FAspVolume   write FAspVolume;
    property    Breathiness : Double  read FBreathiness write FBreathiness;
    property    CrossMix    : Double  read FCrossMix    write FCrossMix;
    property    Modulation  : Boolean read FModulation  write FModulation;
  private
}

    procedure   TSynthesizer.SetSampleRate( aValue : Double);
    begin
      if aValue <> FSampleRate
      then begin
        FSampleRate                := aValue;
        FBandpassFilter.SampleRate := aValue;
        FVocalTract    .SampleRate := aValue;
        FThroat        .SampleRate := aValue;
      end;
    end;


    procedure   TSynthesizer.SetGlotPitch( aValue : Double);
    begin
      if aValue <> FGlotPitch
      then begin
        FGlotPitch := aValue;
        FFrequency := UnitsToFrequency( aValue);
      end;
    end;


//  private

    function    TSynthesizer.GetBandpassBandwidth: Double;
    begin
      Result := FBandPassFilter.Bandwidth;
    end;


    procedure   TSynthesizer.SetBandpassBandwidth( aValue : Double);
    begin
      FBandPassFilter.Bandwidth := aValue;
    end;


    function    TSynthesizer.GetBandpassCenterFrequency: Double;
    begin
      Result := FBandPassFilter.CenterFrequency;
    end;


    procedure   TSynthesizer.SetBandpassCenterFrequency( aValue : Double);
    begin
      FBandPassFilter.CenterFrequency := aValue;
    end;


    function    TSynthesizer.GetThroatCutoff: Double;
    begin
      Result := FThroat.Cutoff;
    end;


    procedure   TSynthesizer.SetThroatCutoff( aValue : Double);
    begin
      FThroat.Cutoff := aValue;
    end;


    function    TSynthesizer.GetThroatGain: Double;
    begin
      Result := FThroat.Gain;
    end;


    procedure   TSynthesizer.SetThroatGain( aValue : Double);
    begin
      FThroat.Gain := aValue;
    end;


    function    TSynthesizer.GetMouthReflectionCoeff: Double;
    begin
      Result := FVocalTract.MouthReflectionCoeff;
    end;


    procedure   TSynthesizer.SetMouthReflectionCoeff( aValue : Double);
    begin
      FVocalTract.MouthReflectionCoeff := aValue;
    end;


    function    TSynthesizer.GetNasalReflectionCoeff: Double;
    begin
      Result := FVocalTract.NasalReflectionCoeff;
    end;


    procedure   TSynthesizer.SetNasalReflectionCoeff( aValue : Double);
    begin
      FVocalTract.NasalReflectionCoeff := aValue;
    end;


//  private

    procedure   TSynthesizer.SetDefaults;
    begin
      GlotPitch               :=  -12.0; // 220 Hz
      GlotVolume              :=    1.0; // Max volume
      AspVolume               :=    0.5;
      Breathiness             :=    2.5;
      BandpassCenterFrequency := 5000.0;
      BandpassBandWidth       :=  250.0;
      CrossMix                :=    0.5;
      Modulation              := True;
      ThroatCutoff            := 1600.0;
      ThroatGain              :=    0.5;
      MouthReflectionCoeff    := 4000.0;
      NasalReflectionCoeff    := 4000.0;
    end;


//  public

    constructor TSynthesizer.Create;
    begin
      FNoise          := TNoise         .Create;
      FNoiseFilter    := TNoiseFilter   .Create;
      FBandPassFilter := TBandPassFilter.Create;
      FVocalTract     := TVocalTract    .Create;
      FThroat         := TThroat        .Create;
      SetDefaults;
    end;


    destructor  TSynthesizer.Destroy; // override;
    begin
      FreeAndNil( FThroat        );
      FreeAndNil( FVocalTract    );
      FreeAndNil( FBandPassFilter);
      FreeAndNil( FNoiseFilter   );
      FreeAndNil( FNoise         );
    end;


    function    TSynthesizer.NextValue( anInput: Double): Double;
    var
      aNoise       : Double;
      aPulsedNoise : Double;
      aPulse       : Double;
      aCrossMix    : Double;
    begin
      aPulse       := anInput;
      aNoise       := FNoiseFilter.NextValue( FNoise.NextValue);
      aPulsedNoise := aNoise * aPulse;
      aPulse       := GlotVolume * ( aPulse * ( 1.0 - Breathiness) + aPulsedNoise * Breathiness);

      if   Modulation
      then begin
        aCrossMix := Min( GlotVolume * CrossMix, 1.0);
        Result    := aPulsedNoise * aCrossMix + ( 1.0 - aCrossMix) * aNoise;
      end
      else Result := aNoise;

      Result := FVocalTract.NextValue(( aPulse + AspVolume * Result) * 0.125 {VT_SCALE}, FBandpassFilter.NextValue( Result));
      Result := Result + FThroat.NextValue( aPulse * 0.125 {VT_SCALE});
      Result := Result * 100;
    end;




// Stuff from syn_structs.h

const

  TOTAL_REGIONS            =     8;
  TOTAL_NASAL_SECTIONS     =     6;
  TABLE_LENGTH             =   512;
  TEMPERATURE_DEF          =    32.0;
  GLOT_PITCH_DEF           =     0.0;
  GLOT_VOL_DEF             =    60;
  CIRC_BUFF_SIZE           =  2048;
  CIRC_BUFF2_SIZE          =  8192;
  TOTAL_SECTIONS           =    10;
  TOTAL_NASAL_COEFFICIENTS = TOTAL_NASAL_SECTIONS;
  TOTAL_COEFFICIENTS       = TOTAL_REGIONS;
  TOTAL_ALPHA_COEFFICIENTS = 3;
  TABLE_MODULUS            = TABLE_LENGTH - 1;


var

  controlPeriod     : Integer;
  sampleRate        : Integer;
  actualTubeLength  : Double;
  breathinessFactor : Double;
  crossmixFactor    : Double;
  dampingFactor     : Double;
  tableDiv1         : Integer;
  tableDiv2         : Integer;
  tnLength          : Double;
  tnDelta           : Double;
  basicIncrement    : Double;
  currentPosition   : Double;
  wavetable         : array[ 0 .. TABLE_LENGTH             - 1  ] of Double;
  nasal_coeff       : array[ 0 .. TOTAL_NASAL_COEFFICIENTS - 1  ] of Double;
  oropharynx_coeff  : array[ 0 .. TOTAL_COEFFICIENTS       - 1] of Double;
  alpha             : array[ 0 .. TOTAL_ALPHA_COEFFICIENTS - 1] of Double;

  nasal             : array[ 0 .. TOTAL_NASAL_SECTIONS - 1, 0 .. 1, 0 .. 1] of Double;
  oropharynx        : array[ 0 .. TOTAL_SECTIONS        -1, 0 .. 1, 0 .. 1] of Double;

  GLogger           : TTubeLogger;
  CircularBuffer    : TCircularBuffer = nil;
  CircularBuffer2   : TCircularBuffer = nil;

const

  // OROPHARYNX REGIONS
  R1 = 0; // S1
  R2 = 1; // S2
  R3 = 2; // S3
  R4 = 3; // S4 & S5
  R5 = 4; // S6 & S7
  R6 = 5; // S8
  R7 = 6; // S9
  R8 = 7; // S10
  // #define TOTAL_REGIONS             8  ( moved to structs.h)

  // OROPHARYNX SCATTERING JUNCTION COEFFICIENTS ( BETWEEN EACH REGION)
  C1 = R1;  // R1-R2 ( S1-S2)
  C2 = R2;  // R2-R3 ( S2-S3)
  C3 = R3;  // R3-R4 ( S3-S4)
  C4 = R4;  // R4-R5 ( S5-S6)
  C5 = R5;  // R5-R6 ( S7-S8)
  C6 = R6;  // R6-R7 ( S8-S9)
  C7 = R7;  // R7-R8 ( S9-S10)
  C8 = R8;  // R8-AIR ( S10-AIR)
  // #define TOTAL_COEFFICIENTS        TOTAL_REGIONS  ( moved to structs.h)

  // OROPHARYNX SECTIONS
  S1  = 0; // R1
  S2  = 1; // R2
  S3  = 2; // R3
  S4  = 3; // R4
  S5  = 4; // R4
  S6  = 5; // R5
  S7  = 6; // R5
  S8  = 7; // R6
  S9  = 8; // R7
  S10 = 9;  // R8
  // #define TOTAL_SECTIONS            10  ( moved to structs.h)

  // NASAL TRACT SECTIONS
  N1 = 0;
  VELUM = N1;
  N2 = 1;
  N3 = 2;
  N4 = 3;
  N5 = 4;
  N6 = 5;
  // #define TOTAL_NASAL_SECTIONS      6 ( moved to structs.h)

  // NASAL TRACT COEFFICIENTS
  NC1 = N1; // N1-N2
  NC2 = N2; // N2-N3
  NC3 = N3; // N3-N4
  NC4 = N4; // N4-N5
  NC5 = N5; // N5-N6
  NC6 = N6; // N6-AIR
  // #define TOTAL_NASAL_COEFFICIENTS  TOTAL_NASAL_SECTIONS ( moved to structs.h)

  // THREE-WAY JUNCTION ALPHA COEFFICIENTS
  LEFT = 0;
  RIGHT = 1;
  UPPER = 2;
  // #define TOTAL_ALPHA_COEFFICIENTS  3

  // FRICATION INJECTION COEFFICIENTS
  FC1 = 0;  // S3
  FC2 = 1;  // S4
  FC3 = 2;  // S5
  FC4 = 3;  // S6
  FC5 = 4;  // S7
  FC6 = 5;  // S8
  FC7 = 6;  // S9
  FC8 = 7;  // S10
  TOTAL_FRIC_COEFFICIENTS = 8;


  // GLOTTAL SOURCE OSCILLATOR TABLE VARIABLES
  // #define TABLE_LENGTH              512
  // #define TABLE_MODULUS             ( TABLE_LENGTH-1)

  // WAVEFORM TYPES
  WPULSE = 0;
  WSINE  = 1;

  // OVERSAMPLING FIR FILTER CHARACTERISTICS
  FIR_BETA   = 0.2;
  FIR_GAMMA  = 0.1;
  FIR_CUTOFF = 0.00000001;

  // PITCH VARIABLES
  PITCH_BASE   = 220.0;
  PITCH_OFFSET = 3; // MIDDLE C = 0

  // RANGE OF ALL VOLUME CONTROLS
  VOL_MAX = 60;

  // SCALING CONSTANT FOR INPUT TO VOCAL TRACT & THROAT ( MATCHES DSP)
  // #define VT_SCALE                  0.03125     /*  2^( -5)  */
  // this is a temporary fix only, to try to match dsp synthesizer
  VT_SCALE = 0.125; // 2^( -3)

  // CONSTANTS FOR THE FIR FILTER
  LIMIT              = 200;
  BETA_OUT_OF_RANGE  = 1;
  GAMMA_OUT_OF_RANGE = 2;
  GAMMA_TOO_SMALL    = 3;

  // CONSTANTS FOR NOISE GENERATOR
  FACTOR       = 377.0;
  INITIAL_SEED = 0.7892347;

  // MAXIMUM SAMPLE VALUE
  RANGE_MAX = 32767.0;

  // FUNCTION RETURN CONSTANTS
  ERROR   = -1;
  SUCCESS =  0;

  // BI-DIRECTIONAL TRANSMISSION LINE POINTERS
  TOP = 0;
  BOTTOM = 1;


  // SAMPLE RATE CONVERSION CONSTANTS
  ZERO_CROSSINGS = 13;                 // SRC CUTOFF FRQ
  LP_CUTOFF      = ( 11.0 / 13.0);     // ( 0.846 OF NYQUIST)

  N_BITS         = 16;
  L_BITS         = 8;
  L_RANGE        = 256;                // must be 2^L_BITS
  M_BITS         = 8;
  M_RANGE        = 256;                // must be 2^M_BITS
  FRACTION_BITS  = ( L_BITS + M_BITS);
  FRACTION_RANGE = 65536;              // must be 2^FRACTION_BITS
  FILTER_LENGTH  = ( ZERO_CROSSINGS * L_RANGE);
  FILTER_LIMIT   = ( FILTER_LENGTH - 1);

  N_MASK        = $FFFF0000;
  L_MASK        = $0000FF00;
  M_MASK        = $000000FF;
  FRACTION_MASK = $0000FFFF;

  BETA         = 5.658; // kaiser window parameters
  IzeroEPSILON = 1E-21;

  OUTPUT_SRATE_LOW  = 22050.0;  // not used apparently
  OUTPUT_SRATE_HIGH = 44100.0;  // not used apparently
  BUFFER_SIZE       = 1024;     // ring buffer size

  // SIZE IN BITS PER OUTPUT SAMPLE
  BITS_PER_SAMPLE = 16;         // not used apparently


var

  // REFLECTION AND RADIATION FILTER MEMORY
  a10, b11, a20, a21, b21: Double;

  // NASAL REFLECTION AND RADIATION FILTER MEMORY
  na10, nb11, na20, na21, nb21: Double;

  // THROAT LOWPASS FILTER MEMORY, GAIN
  tb1, ta0, throatGain: Double;

  // FRICATION BANDPASS FILTER MEMORY
  bpAlpha, bpBeta, bpGamma: Double;


const

  // TEMPORARY SAMPLE STORAGE VALUES
  maximumSampleValue : Double  = 0.0;
  numberSamples      : LongInt = 0;

type

  TRadii        = array[ 0 .. TOTAL_REGIONS - 1] of Double;
  TCoefficients = array[ 0 .. LIMIT            ] of Double; // Yes, no -1 there
  TDoubles      = array[ 0 .. MaxInt div SizeOf( Double) - 1] of Double;
  PDoubles      = ^TDoubles;


var

  // MEMORY FOR FRICATION TAPS
  fricationTap: array[ 0 .. TOTAL_FRIC_COEFFICIENTS - 1] of Double;

  // VARIABLES FOR FIR LOWPASS FILTER
  FIRData, FIRCoef    : PDoubles;
  FIRPtr , numberTaps : Integer;

  // VARIABLES FOR SAMPLE RATE CONVERSION

  sampleRateRatio : Double;
  h               : array[ 0 .. FILTER_LENGTH - 1] of Double;
  deltaH          : array[ 0 .. FILTER_LENGTH  -1] of Double;
  buffer          : array[ 0 .. BUFFER_SIZE   - 1] of Double;

  fillPtr : Integer;

const

  emptyPtr     : Integer  = 0;
  timeRegister : Cardinal = 0;

var

  padSize, fillSize: Integer;

  timeRegisterIncrement, filterIncrement, phaseIncrement: Cardinal;

  originalTime, outputTime, signal1, signal2: Double;  // Keep track of where we are in resampling
  originalPeriod, outputPeriod: Double;                // The two sample periods
  sampleCount: Integer;


type

  _postureRateParameters = record
    glotPitch      : Double;  //  -0.0
    glotPitchDelta : Double;  //   0
    glotVol        : Double;  //  60
    glotVolDelta   : Double;  //   0
    aspVol         : Double;  //   0
    aspVolDelta    : Double;  //   0
    fricVol        : Double;  //   0
    fricVolDelta   : Double;  //   0
    fricPos        : Double;  //   0
    fricPosDelta   : Double;  //   0
    fricCF         : Double;  // 5000
    fricCFDelta    : Double;  //   0
    fricBW         : Double;  // 250
    fricBWDelta    : Double;  //   0
    radius         : TRadii;  //   0.8, 1.67, 1.905, 1.985, 0.81, 0.495, 0.73, 1.485
    radiusDelta    : TRadii;  //   0,0,0,0,0,0,0,0
    velum          : Double;  //   0
    velumDelta     : Double;  //   0
  end;
  T_postureRateParameters = _postureRateParameters;


var

// _postureRateParameters current := begin -0.0, 0, 60, 0, 0, 0, 0, 0, 8, 0, 5000, 0, 250, 0, 0.8, 1.67, 1.905, 1.985, 0.81, 0.495, 0.73, 1.485, 0,0,0,0,0,0,0,0, 0,0);  // "ee"

  current : _postureRateParameters = ( // "ee"
    glotPitch      :   -0.0;
    glotPitchDelta :    0;
    glotVol        :   60;
    glotVolDelta   :    0;
    aspVol         :    0;
    aspVolDelta    :    0;
    fricVol        :    0;
    fricVolDelta   :    0;
    fricPos        :    8;
    fricPosDelta   :    0;
    fricCF         : 5000;
    fricCFDelta    :    0;
    fricBW         :  250;
    fricBWDelta    :    0;
    radius         : ( 0.8, 1.67, 1.905, 1.985, 0.81, 0.495, 0.73, 1.485);
    radiusDelta    : ( 0, 0, 0, 0, 0, 0, 0, 0);
    velum          : 0;
    velumDelta     : 0
  );


  originalDefaults : _postureRateParameters = ( // "ee"
    glotPitch      : GLOT_PITCH_DEF;
    glotPitchDelta :    0.1;
    glotVol        : GLOT_VOL_DEF;
    glotVolDelta   :    0.1;
    aspVol         :    0;
    aspVolDelta    :    0.1;
    fricVol        :    0;
    fricVolDelta   :    0.1;
    fricPos        :    8;
    fricPosDelta   :    0.1;
    fricCF         : 5000;
    fricCFDelta    :   10;
    fricBW         :  250;
    fricBWDelta    :    0;
    radius         : ( 0.8, 1.67, 1.905, 1.985, 0.81, 0.495, 0.73, 1.485);
    radiusDelta    : ( 0, 0, 0, 0, 0, 0, 0, 0);
    velum          : 0;
    velumDelta     : 0
  );



  // 18 of the utterance rate parameters follow in the next 21 lines There are 6 more
  // static void *currentPointer = &current;

  // Assign values to utterance rate parameters.There are 25 total.
  // waveform -- 0 or 1 -- selects pulse or sin wave, and controlRate selects the input control rate -- 1 to 1000
  // the default is 100, i.e. every 10 millisecond. waveform and controlRate really don't need to be archived.
  apScale                    : Double = 2.5;
  balance                    : Double = 0;
  breathiness                : Double = 2.5;
  channels                   : Integer = 2;
  controlRate                : Double = 100;                            // ****

  tubeLength                 : Double = 17;
  lossFactor                 : Double = 0.0;
  mixOffset                  : Double = 48.0;
  modulation                 : Integer = 1;
  mouthCoef                  : Double = 4000.0;
  noseCoef                   : Double = 4000.0;
  noseRadiusOriginalDefaults : array [ 0  .. TOTAL_NASAL_SECTIONS - 1] of Double = ( 1.35, 1.35, 1.7, 1.7, 1.3, 0.9);  // fixed nose radii ( 0 - 3 cm)

  temperature                : Double = 32;      // tube temperature ( 25 - 40 C)
  throatCutoff               : Double = 1500.0;  // throat lp cutoff ( 50 - nyquist Hz)
  throatVol                  : Double = 6.0;     // throat volume ( 0 - 48 dB)
  tnMax                      : Double = 40;      // % glottal pulse fall time maximum
  tnMin                      : Double = 16;      // % glottal pulse fall time minimum
  tp                         : Double = 35;      // % glottal pulse rise time
  volume                     : Double = 60;      // master volume ( 0 - 60 dB)
  waveform                   : Integer = 0;      // GS waveform type ( 0=PULSE, 1=SINE)

  current_ptr                : Integer = 1;
  noseRadius                 : array[ 0 ..TOTAL_NASAL_SECTIONS - 1] of Double = ( 1.35, 1.35, 1.7, 1.7, 1.3, 0.9);  // fixed nose radii ( 0 - 3 cm)

  outputRate                 : Double = 44100;   // output sample rate ( 22.05, 44.1 KHz)
  prev_ptr                   : Integer = 0;
  run                        : Integer = 0;

  // Flag to signal pitch period

  pitchFlag                  : Integer = 0;      // 0 unless pitch period just started
  initialized                : Boolean = False;






// local globals .. erm ... function bound statics from C

  seed             : Double  = INITIAL_SEED;
  noiseX           : Double  = 0.0;
  reflectionY      : Double  = 0.0;
  radiationX       : Double  = 0.0;
  radiationY       : Double  = 0.0;
  nasalReflectionY : Double  = 0.0;
  nasalRadiationX  : Double  = 0.0;
  nasalRadiationY  : Double  = 0.0;
  throatY          : Double  = 0.0;
  xn1              : Double  = 0.0;
  xn2              : Double  = 0.0;
  yn1              : Double  = 0.0;
  yn2              : Double  = 0.0;
  fillCounter      : Integer = 0;


  value  : Double;
  value1 : Integer;


  // GLOBAL FUNCTIONS ( LOCAL TO THIS FILE)  **********************************

  function  initializeSynthesizer: Integer;                                                                     forward;
  procedure initializeWavetable;                                                                                forward;
  function  speedOfSound( temperature: Double): Double;                                                         forward;
  procedure updateWavetable( amplitude: double);                                                                forward;
  procedure initializeFIR( beta: Double; gamma: Double; cutoff: Double);                                        forward;
  function  noise: Double;                                                                                      forward;
  function  noiseFilter( input: Double): Double;                                                                forward;
  procedure initializeMouthCoefficients( coeff: Double);                                                        forward;
  function  reflectionFilter( input: Double): Double;                                                           forward;
  function  radiationFilter( input: Double): Double;                                                            forward;
  procedure initializeNasalFilterCoefficients( coeff: Double);                                                  forward;
  function  nasalReflectionFilter( input: Double): Double;                                                      forward;
  function  nasalRadiationFilter( input: Double): Double;                                                       forward;

  procedure initializeNasalCavity;                                                                              forward;
  procedure initializeThroat;                                                                                   forward;
  procedure calculateTubeCoefficients;                                                                          forward;
  procedure setFricationTaps;                                                                                   forward;
  procedure calculateBandpassCoefficients;                                                                      forward;
  function  mod0( value: Double): Double;                                                                       forward;
  procedure incrementTablePosition( frequency: Double);                                                         forward;
  function  oscillator( frequency: Double): Double;                                                             forward;
  function  vocalTract( input: Double; frication: Double): Double;                                              forward;
  function  throat( input: Double): Double;                                                                     forward;
  function  bandpassFilter( input: Double): Double;                                                             forward;
  function  amplitude( decibelLevel: Double): Double;                                                           forward;
  function  frequency( pitch: Double): Double;                                                                  forward;
  function  maximallyFlat( beta: Double; gamma: Double; var np: Integer; coefficient : PDoubles): Integer;      forward;
  procedure trim( cutoff: Double; var numberCoefficients: Integer; coefficient: PDoubles);                      forward;
  procedure rationalApproximation( number: Double; var order, numerator, denominator: Integer);                 forward;
  function  FIRFilter( input: Double; needOutput: Integer): Double;                                             forward;
  function  increment( aPointer: Integer; modulus: Integer): Integer;                                           forward;
  function  decrement( aPointer: Integer; modulus: Integer): Integer;                                           forward;
  procedure initializeConversion;                                                                               forward;
  procedure initializeFilter;                                                                                   forward;
  function  Izero2( x: Double): Double;                                                                         forward;
  procedure initializeBuffer;                                                                                   forward;
  procedure dataFill( data: Double);                                                                            forward;
  procedure dataEmpty;                                                                                          forward;
  procedure putCircBuff2( value: Double);                                                                       forward;
  procedure flushBuffer;                                                                                        forward;
  procedure srIncrement( var aPointer: Integer; modulus: Integer);                                              forward;
  procedure srDecrement( var aPointer: Integer; modulus: Integer);                                              forward;

  procedure initCircBuff;                                                                                       forward;
  procedure initCircBuff2;                                                                                      forward;


  // ////////////////////////////////////////////////////////////////////////////////////////////////////////////////// /

  procedure Log( const aMsg: string);
  begin
    if   Assigned( GLogger)
    then GLogger( aMsg);
  end;


  // ////////////////////////////////////////////////////////////////////////////////////////////////////////////////// /

  procedure LogFmt( const aFmt: string; const anArgs: array of const);
  begin
    Log( Format( aFmt, anArgs));
  end;


  // const nValue(( ( x) and N_MASK) shr FRACTION_BITS);
  function nValue( x: Cardinal): Cardinal; begin Result := ( x and N_MASK) shr FRACTION_BITS; end;

  // const lValue(( ( x) and L_MASK) shr M_BITS);
  function lValue( x: Cardinal): Cardinal; begin Result := ( x and L_MASK) shr M_BITS; end;

  // const mValue(( x) and M_MASK);
  function mValue( x: Cardinal): Cardinal; begin Result := x and M_MASK; end;

  // const fractionValue(( x) and FRACTION_MASK);
  function fractionValue( x: Cardinal): Cardinal; begin Result := x and FRACTION_MASK; end;

  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  speedOfSound
  //
  //  purpose:  Returns the speed of sound according to the value of
  //                        the temperature ( in Celsius degrees).
  //
  //        arguments:      temperature
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function speedOfSound( temperature: Double): Double;
  begin
    Result := 331.4 + ( 0.6 * temperature);
    // printf( 'In tube.c:493 ( sp-of-snd) temperature passed is %f computed value s-o-snd %f\n', temperature, computedSpeed);
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  initializeSynthesizer
  //
  //  purpose:  Initializes all variables so that the synthesis can
  //                        be run.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  speedOfSound, amplitude, initializeWavetable,
  //                        initializeFIR, initializeNasalFilterCoefficients,
  //                        initializeNasalCavity, initializeThroat,
  //                        initializeConversion
  //
  //  aLibrary
  //  functions:  Round
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function initializeSynthesizer: Integer;
  var
    nyquist : Double;
    c       : Double;
  begin
    // printf( 'tube.c:531 Initialising synthesisern');

    if   not initialized
    then begin
      initCircBuff;
      initCircBuff2;
      originalTime := 0;
      outputTime   := 0;
      sampleCount  := 0;
      initialized  := True;
    end;


    // CALCULATE THE SAMPLE RATE, BASED ON NOMINAL TUBE LENGTH AND SPEED OF SOUND

    if   tubeLength > 0.0
    then begin
      c := speedOfSound( temperature);
      controlPeriod := Round(( c * TOTAL_SECTIONS * 100.0) /( tubeLength * controlRate));
      // printf( 'tube.c:530 ControlPeriod is %d n', controlPeriod);  // rint(( c * TOTAL_SECTIONS * 100.0) /( tubeLength * controlRate))); // (( double  getControlPeriod));
      sampleRate := Round( controlRate * controlPeriod);       // ****
      originalPeriod := 1.0 / sampleRate;          // ****

      // printf( 'tube.c:555 SampleRate is %f control period is %d control rate is %f n', controlRate * controlPeriod, controlPeriod, controlRate);  // sampleRate);
      actualTubeLength := ( c * TOTAL_SECTIONS * 100.0) / sampleRate;
      // printf( 'tube.c:557 Actual tube length is %f originalPeriod is %f sampleRate is %un', actualTubeLength, originalPeriod, sampleRate);
      nyquist := 0.5 * sampleRate;
    end
    else begin
      // fprintf( stderr, 'tube.c:538Illegal tube length.n');
      Result := ERROR;
      Exit;
    end;

    // CALCULATE THE BREATHINESS FACTOR
    breathinessFactor := breathiness / 100.0;

    // CALCULATE CROSSMIX FACTOR
    // printf( 'mixOffset is %f\n', mixOffset);
    crossmixFactor := 1.0 / amplitude( mixOffset);
    // printf( '**** mixOffset Factor is %f\n', crossmixFactor);

    // CALCULATE THE DAMPING FACTOR
    dampingFactor := ( 1.0 - ( lossFactor / 100.0));
    // printf( 'tube.c:563 dampingFactor is %f, lossFactor is %f\n', dampingFactor, lossFactor);

    // INITIALIZE THE WAVE TABLE
    initializeWavetable;

    // INITIALIZE THE FIR FILTER
    initializeFIR( FIR_BETA, FIR_GAMMA, FIR_CUTOFF);

    // INITIALIZE REFLECTION AND RADIATION FILTER COEFFICIENTS FOR MOUTH
    initializeMouthCoefficients(( nyquist - mouthCoef) / nyquist);

    // INITIALIZE REFLECTION AND RADIATION FILTER COEFFICIENTS FOR NOSE
    initializeNasalFilterCoefficients(( nyquist - noseCoef) / nyquist);

    // INITIALIZE NASAL CAVITY FIXED SCATTERING COEFFICIENTS
    initializeNasalCavity;

    // INITIALIZE THE THROAT LOWPASS FILTER
    initializeThroat;

    // INITIALIZE THE SAMPLE RATE CONVERSION ROUTINES
    initializeConversion;

    temperature := TEMPERATURE_DEF;

    // printf( 'tube.c:579 SampleRate is %f control period is %d control rate is %f n', controlRate * controlPeriod, controlPeriod, controlRate);  // sampleRate);

    outputPeriod := 1.0 / outputRate;

    // Return success

    Result := SUCCESS;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  initializeWavetable
  //
  //  purpose:  Calculates the initial glottal pulse and stores it
  //                        in the wavetable, for use in the oscillator.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  Round
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure initializeWavetable;
  var
    i, j : Integer;
    x    : Double;
    x2   : Double;
    x3   : Double;
  begin
    // CALCULATE WAVE TABLE PARAMETERS

    tableDiv1 := Round( TABLE_LENGTH * ( tp / 100.0));
    tableDiv2 := Round( TABLE_LENGTH * (( tp + tnMax) / 100.0));  // **** works for tnMax but not if tnMin is substituted?

    // printf( 'tableDiv1 is %d tableDiv2 is %dn', tableDiv1, tableDiv2);

    tnLength := tableDiv2 - tableDiv1;
    tnDelta := Round( TABLE_LENGTH * (( tnMax - tnMin) / 100.0));
    basicIncrement := TABLE_LENGTH / sampleRate;
    currentPosition := 0;

    // INITIALIZE THE WAVETABLE WITH EITHER A GLOTTAL PULSE OR SINE TONE
    if waveform = WPULSE
    then begin
      // CALCULATE RISE PORTION OF WAVE TABLE
      for i := 0 to tableDiv1 - 1
      do begin
        x := i / tableDiv1;
        x2 := x  * x;
        x3 := x2 * x;
        wavetable[ i] := ( 3.0 * x2) - ( 2.0 * x3);
      end;

      // CALCULATE FALL PORTION OF WAVE TABLE
      j := 0;

      for i := tableDiv1 to tableDiv2 -1
      do begin
        x := j / tnLength;
        wavetable[ i] := 1.0 - ( x * x);
        Inc( j);
      end;

      // SET CLOSED PORTION OF WAVE TABLE
      for i := tableDiv2 to TABLE_LENGTH - 1
      do wavetable[ i] := 0.0;
    end
    else begin
      // SINE WAVE
      for i := 0 to TABLE_LENGTH - 1
      do wavetable[ i] := sin(( i / TABLE_LENGTH) * 2.0 * PI );
    end;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  updateWavetable
  //
  //  purpose:  Rewrites the changeable part of the glottal pulse
  //                        according to the amplitude.
  //
  //        arguments:      amplitude
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  rint
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure updateWavetable( amplitude: Double);
  var
    i, j        : Integer;
    x           : Double;
    newDiv2     : Double;
    newTnLength : Double;
  begin
    // CALCULATE NEW CLOSURE POINT, BASED ON AMPLITUDE
    newDiv2     := tableDiv2 - round( amplitude * tnDelta);
    newTnLength := newDiv2 - tableDiv1;
    // printf( "Values in updateWavetable tube.c:684 for i, %d newDiv2 %d, and tableDiv2 %d are:", i, newDiv2, tableDiv2);

    // RECALCULATE THE FALLING PORTION OF THE GLOTTAL PULSE
    j := 0;
    i := TableDiv1;

    while i < newDiv2
    do begin
      x := j / newTnLength;
      wavetable[ i] := 1.0 - ( x * x);
      inc( i);
      inc( j);
    end;

    // FILL IN WITH CLOSED PORTION OF GLOTTAL PULSE
    i := Round( newDiv2);

    while i < tableDiv2
    do begin
      wavetable[ i] := 0.0;
      Inc( i);
    end;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  initializeFIR
  //
  //  purpose:  Allocates memory and initializes the coefficients
  //                        for the FIR filter used in the oversampling oscillator.
  //
  //        arguments:      beta, gamma, cutoff
  //
  //  internal
  //  functions:  maximallyFlat, trim
  //
  //  aLibrary
  //  functions:  calloc
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure initializeFIR( beta: Double; gamma: Double; cutoff: Double);
  var
    i, aPointer, increment, numberCoefficients : Integer;
    coefficient : TCoefficients;
  begin
    // DETERMINE IDEAL LOW PASS FILTER COEFFICIENTS
    maximallyFlat( beta, gamma, numberCoefficients, @ coefficient);

    // TRIM LOW-VALUE COEFFICIENTS
    trim( cutoff, numberCoefficients, @ coefficient);

    // DETERMINE THE NUMBER OF TAPS IN THE FILTER
    numberTaps := ( numberCoefficients * 2) - 1;

    // ALLOCATE MEMORY FOR DATA AND COEFFICIENTS
  // FIRData := ( Double )calloc( numberTaps, SizeOf( Double));
  // FIRCoef := ( Double )calloc( numberTaps, SizeOf( Double));

    GetMem( FIRData, numberTaps * SizeOf( Double));
    GetMem( FIRCoef, numberTaps * SizeOf( Double));

    // INITIALIZE THE COEFFICIENTS
    increment := -1;
    aPointer  := numberCoefficients;

    for i := 0 to numberTaps - 1
    do begin
      FIRCoef^[ i] := coefficient[ aPointer];
      aPointer := aPointer + increment;

      if   aPointer <= 0
      then begin
        aPointer := 2;
        increment := 1;
      end;
    end;

    // SET POINTER TO FIRST ELEMENT
    FIRPtr := 0;

  {$ifdef DEBUG}
    // PRINT OUT
    // printf( 'n');
    // for( i := 0; i < numberTaps; i++ )
    //  printf( 'FIRCoef[ %-d] := %11.8fn', i, FIRCoef[ i]);
  {$endif}
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  noise
  //
  //  purpose:  Returns one value of a random sequence.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function noise: Double;
  var
    product : Double;
  begin
    product := seed * FACTOR;
    seed    := product - Floor( product);
    Result  := ( seed - 0.5);
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  noiseFilter
  //
  //  purpose:  One-zero lowpass filter.
  //
  //        arguments:      input
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function noiseFilter( input: Double): Double;
  begin
    Result := input + noiseX;
    noiseX := input;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  initializeMouthCoefficients
  //
  //  purpose:  Calculates the reflection/radiation filter coefficients
  //                        for the mouth, according to the mouth aperture
  //                        coefficient.
  //
  //        arguments:      coeff - mouth aperture coefficient
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  fabs
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure initializeMouthCoefficients( coeff: Double);
  begin
    b11 := -coeff;
    a10 := 1.0 - Abs( b11);
    a20 := coeff;
    a21 := -a20;
    b21 := -a20;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  reflectionFilter
  //
  //  purpose:  Is a variable, one-pole lowpass filter, whose cutoff
  //                        is determined by the mouth aperture coefficient.
  //
  //        arguments:      input
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function reflectionFilter( input: Double): Double;
  begin
    Result      := ( a10 * input) - ( b11 * reflectionY);
    reflectionY := Result;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  radiationFilter
  //
  //  purpose:  Is a variable, one-zero, one-pole, highpass filter,
  //                        whose cutoff point is determined by the mouth aperture
  //                        coefficient.
  //
  //        arguments:      input
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function radiationFilter( input: Double): Double;
  begin
    Result     := ( a20 * input) + ( a21 * radiationX) - ( b21 * radiationY);
    radiationX := input;
    radiationY := Result;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  initializeNasalFilterCoefficients
  //
  //  purpose:  Calculates the fixed coefficients for the nasal
  //                        reflection/radiation filter pair, according to the
  //                        nose aperture coefficient.
  //
  //        arguments:      coeff - nose aperture coefficient
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  fabs
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure initializeNasalFilterCoefficients( coeff: Double);
  begin

    nb11 := -coeff;
    na10 := 1.0 - abs( nb11);

    na20 := coeff;
    na21 := -na20;
    nb21 := -na20;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  nasalReflectionFilter
  //
  //  purpose:  Is a one-pole lowpass filter, used for terminating
  //                        the end of the nasal cavity.
  //
  //        arguments:      input
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function nasalReflectionFilter( input: Double): Double;
  begin
    Result           := ( na10 * input) - ( nb11 * nasalReflectionY);
    nasalReflectionY := Result;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  nasalRadiationFilter
  //
  //  purpose:  Is a one-zero, one-pole highpass filter, used for the
  //                        radiation characteristic from the nasal cavity.
  //
  //        arguments:      input
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function nasalRadiationFilter( input: Double): Double;
  begin
    Result := ( na20 * input) + ( na21 * nasalRadiationX) - ( nb21 * nasalRadiationY);
    nasalRadiationX := input;
    nasalRadiationY := Result;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  synthesize
  //
  //  purpose:  Performs the actual synthesis of sound samples.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  setControlRateParameters, frequency, amplitude,
  //                        calculateTubeCoefficients, noise, noiseFilter,
  //                        updateWavetable, oscillator, vocalTract, throat,
  //                        dataFill, sampleRateInterpolation
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure synthesize;
  var
    f0, ax, ah1, pulse, lp_noise, pulsed_noise, crossmix, gsignal: Double;  // signal,
  begin
    // SAMPLE RATE LOOP TO FILL BUFFER FOR Core Audio IOProc

    while True
    do begin

      // CONVERT PARAMETERS HERE
      f0 := frequency( current.glotPitch);
      ax := amplitude( current.glotVol  );


      // if j == 10
      // then printf( "current.glotVol is %f, ax is %f, j is %d", current.glotVol, ax, j);

      ah1 := amplitude( current.aspVol);
      // printf( "Current ah1 is %f", current.aspVol);
      calculateTubeCoefficients;
      setFricationTaps;
      calculateBandpassCoefficients;

      // DO SYNTHESIS HERE
      // CREATE LOW-PASS FILTERED NOISE
      lp_noise := noiseFilter( noise);

      // UPDATE THE SHAPE OF THE GLOTTAL PULSE, IF NECESSARY

      if   waveform = WPULSE
      then updateWavetable( ax);


      // CREATE GLOTTAL PULSE ( OR SINE TONE) by sampling wavetable in oscillator
      pulse := oscillator( f0);

      // CREATE PULSED NOISE
      pulsed_noise := lp_noise * pulse;

      // CREATE NOISY GLOTTAL PULSE
      pulse := ax * (( pulse * ( 1.0 - breathinessFactor)) + ( pulsed_noise * breathinessFactor));

      // CROSS-MIX PURE NOISE WITH PULSED NOISE
      if   modulation <> 0
      then begin
        crossmix := ax * crossmixFactor;

        // crossmix := ( crossmix < 1.0) ? crossmix : 1.0;
        if   crossmix > 1.0
        then crossmix := 1.0;

        gsignal := ( pulsed_noise * crossmix) + ( lp_noise * ( 1.0 - crossmix));
      end
      else gsignal := lp_noise;

      // PUT SIGNAL THROUGH VOCAL TRACT
      gsignal := vocalTract(( ( pulse + ( ah1 * gsignal)) * VT_SCALE), bandpassFilter( gsignal));

      // PUT PULSE THROUGH THROAT
      gsignal := gsignal + throat( pulse * VT_SCALE);

      // printf( "gsignal b4 %f     ", gsignal);

      gsignal := gsignal * 100;

      // RESAMPLE SUCCESSIVE VALUES FROM TUBE SAMPLE RATE TO OUTPUT SAMPLE RATE
      dataFill( gsignal);

      originalTime:= originalTime + originalPeriod;
    end;

    // return ( NULL);
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  sampleRateInterpolation
  //
  //  purpose:  Interpolates table values at the sample rate.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure sampleRateInterpolation;
  var
    i : Integer;
  begin
    current.glotPitch := current.glotPitch + current.glotPitchDelta;
    current.glotVol   := current.glotVol   + current.glotVolDelta;
    current.aspVol    := current.aspVol    + current.aspVolDelta;
    current.fricVol   := current.fricVol   + current.fricVolDelta;
    current.fricPos   := current.fricPos   + current.fricPosDelta;
    current.fricCF    := current.fricCF    + current.fricCFDelta;
    current.fricBW    := current.fricBW    + current.fricBWDelta;

    for i := 0 to TOTAL_REGIONS - 1
    do current.radius[ i] := current.radius[ i] + current.radiusDelta[ i];

    current.velum := current.velum + current.velumDelta;

    // printf( "current radius R5 is %f", current.radius[ 4]);
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  initializeNasalCavity
  //
  //  purpose:  Calculates the scattering coefficients for the fixed
  //                        sections of the nasal cavity.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure initializeNasalCavity;
  var
    i, j : Integer;
    radA2, radB2 : Double;
  begin
    // CALCULATE COEFFICIENTS FOR INTERNAL FIXED SECTIONS OF NASAL CAVITY;
    j := NC2;

    for i := N2 to N6 - 1
    do begin
      radA2          := noseRadius[ i]    * noseRadius[ i];
      radB2          := noseRadius[ i+1]  * noseRadius[ i+1];
      nasal_coeff[ j] := ( radA2 - radB2) / ( radA2 + radB2);
      Inc( j);
    end;

    // CALCULATE THE FIXED COEFFICIENT FOR THE NOSE APERTURE
    radA2 := noseRadius[ N6] * noseRadius[ N6];
    radB2 := apScale * apScale;
    nasal_coeff[ NC6] := ( radA2 - radB2) / ( radA2 + radB2);
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  initializeThroat
  //
  //  purpose:  Initializes the throat lowpass filter coefficients
  //                        according to the throatCutoff value, and also the
  //                        throatGain, according to the throatVol value.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  fabs
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure initializeThroat;
  begin
    ta0 := ( throatCutoff * 2.0)/sampleRate;
    tb1 := 1.0 - ta0;

    throatGain := amplitude( throatVol);
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  calculateTubeCoefficients
  //
  //  purpose:  Calculates the scattering coefficients for the vocal
  //                        tract according to the current radii.  Also calculates
  //                        the coefficients for the reflection/radiation filter
  //                        pair for the mouth and nose.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure calculateTubeCoefficients;
  var
    i : Integer;
    radA2, radB2, r0_2, r1_2, r2_2, sum: Double;
  begin
    // CALCULATE COEFFICIENTS FOR THE OROPHARYNX;
    for i := 0 to TOTAL_REGIONS - 2
    do begin
      radA2 := current.radius[ i] * current.radius[ i];
      radB2 := current.radius[ i+1] * current.radius[ i+1];
      oropharynx_coeff[ i] := ( radA2 - radB2) / ( radA2 + radB2);
    end;

    // printf( "Current radius 8 is %f", current.radius[ 7]);
    // CALCULATE THE COEFFICIENT FOR THE MOUTH APERTURE
    radA2 := current.radius[ R8] * current.radius[ R8];
    radB2 := apScale * apScale;
    oropharynx_coeff[ C8] := ( radA2 - radB2) / ( radA2 + radB2);

    // CALCULATE ALPHA COEFFICIENTS FOR 3-WAY JUNCTION
    // NOTE:  SINCE JUNCTION IS IN MIDDLE OF REGION 4, r0_2 = r1_2
    r0_2 := current.radius[ R4] * current.radius[ R4];
    r1_2 := r0_2;
    r2_2 := current.velum * current.velum;
    sum := 2.0 / ( r0_2 + r1_2 + r2_2);
    alpha[ LEFT] := sum * r0_2;
    alpha[ RIGHT] := sum * r1_2;
    alpha[ UPPER] := sum * r2_2;

    // AND 1ST NASAL PASSAGE COEFFICIENT
    radA2 := current.velum * current.velum;
    // printf( "current.velum is %f", current.velum);
    radB2 := noseRadius[ N2] * noseRadius[ N2];
    nasal_coeff[ NC1] := ( radA2 - radB2) / ( radA2 + radB2);
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  setFricationTaps
  //
  //  purpose:  Sets the frication taps according to the current
  //                        position and amplitude of frication.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure setFricationTaps;
  var
      i, integerPart : Integer;
      complement, remainder : Double;
      fricationAmplitude : Double;
  begin
    fricationAmplitude := 10 * ( amplitude( current.fricVol)); // Volume x 10 to be audible. Investigate

    // printf( "tube.c:1329 frication amplitude is %f, current.fricvol %f", fricationAmplitude, current.fricVol);
    // printf( "tube.c:1329 fricPos is %f", current.fricPos);

    // CALCULATE POSITION REMAINDER AND COMPLEMENT
    integerPart := Floor( current.fricPos);
    complement  := current.fricPos - integerPart;
    remainder   := 1.0 - complement;
    // printf( "tube.c:1336 complement is %f, remainder is %f", complement, remainder);

    // SET THE FRICATION TAPS
    for i := FC1 to TOTAL_FRIC_COEFFICIENTS - 1
    do begin
      if i = integerPart
      then begin
        fricationTap[ i] := remainder * fricationAmplitude;

        if   ( i + 1) < TOTAL_FRIC_COEFFICIENTS
        then fricationTap[ ++i] := complement * fricationAmplitude;
      end
      else fricationTap[ i] := 0.0;
    end;

  {$ifdef DEBUG}
    // PRINT OUT
  // printf( 'fricationTaps:  ');
  // for( i := FC1; i < TOTAL_FRIC_COEFFICIENTS; i++ )
  //   printf( '%.6f  ', fricationTap[ i]);
  // printf( 'n');
  {$endif}
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  calculateBandpassCoefficients
  //
  //  purpose:  Sets the frication bandpass filter coefficients
  //                        according to the current center frequency and
  //                        bandwidth.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  tan, cos
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure calculateBandpassCoefficients;
  var
    tanValue, cosValue: Double;
  begin
    tanValue := tan(( PI * current.fricBW) / sampleRate);
    cosValue := cos(( 2.0 * PI * current.fricCF) / sampleRate);

    bpBeta  := ( 1.0 - tanValue) / ( 2.0 * ( 1.0 + tanValue));
    bpGamma := ( 0.5 + bpBeta) * cosValue;
    bpAlpha := ( 0.5 - bpBeta) / 2.0;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  mod0
  //
  //  purpose:  Returns the modulus of 'value', keeping it in the
  //                        range 0 ^. TABLE_MODULUS.
  //
  //        arguments:      value
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function mod0( value: Double): Double;
  begin
    if   value > TABLE_MODULUS
    then value := value - TABLE_LENGTH;

    Result := value;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  incrementTablePosition
  //
  //  purpose:  Increments the position in the wavetable according to
  //                        the desired frequency.
  //
  //        arguments:      frequency
  //
  //  internal
  //  functions:  mod0
  //
  //  aLibrary
  //  functions:  none
  //
  //  Crafty use of mod0 to wrap the currentPosition around at about 510
  //  to keep in bounds of wavetable while repeating.  mod0( value) Returns the
  //  modulus of 'value', keeping it in the range 0 -> TABLE_MODULUS.
  //
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure incrementTablePosition( frequency: Double);
  var
    temp : Double;
  begin
    temp            := currentPosition;
    currentPosition := mod0( currentPosition + ( frequency * basicIncrement));

    if   currentPosition < temp
    then pitchFlag := 1
    else pitchFlag := 0;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  oscillator
  //
  //  purpose:  Is a 2X oversampling interpolating wavetable
  //                        oscillator.
  //
  //        arguments:      frequency
  //
  //  internal
  //  functions:  incrementTablePosition, mod0, FIRFilter
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

{$ifdef OVERSAMPLING_OSCILLATOR}

  function oscillator( frequency: Double): Double;                // 2X OVERSAMPLING OSCILLATOR
  var
    i, lowerPosition, upperPosition : Integer;
    interpolatedValue: Double;
  begin
    for i := 0 to 1
    do begin
      // FIRST INCREMENT THE TABLE POSITION, DEPENDING ON FREQUENCY
      incrementTablePosition( frequency / 2.0);          // ****

      // FIND SURROUNDING INTEGER TABLE POSITIONS
      lowerPosition := Floor( currentPosition);
      upperPosition := Floor( mod0( 1 * lowerPosition + 1));

      // CALCULATE INTERPOLATED TABLE VALUE
      interpolatedValue := ( wavetable[ lowerPosition] + (( currentPosition - lowerPosition) * ( wavetable[ upperPosition] - wavetable[ lowerPosition])));

      // PUT VALUE THROUGH FIR FILTER
      Result := FIRFilter( interpolatedValue, i);
    end;

    // SINCE WE DECIMATE, TAKE ONLY THE SECOND OUTPUT VALUE
  end;

{$else}

  function oscillator( frequency: Double): Double;                // PLAIN OSCILLATOR
  var
    lowerPosition, upperPosition: Integer;
  begin
    // FIRST INCREMENT THE TABLE POSITION, DEPENDING ON FREQUENCY
    incrementTablePosition( frequency);

    // FIND SURROUNDING INTEGER TABLE POSITIONS
    lowerPosition := Floor( currentPosition);
    upperPosition := Floor( mod0( lowerPosition + 1));

  // RETURN INTERPOLATED TABLE VALUE
    Result := ( wavetable[ lowerPosition] + (( currentPosition - lowerPosition) * ( wavetable[ upperPosition] - wavetable[ lowerPosition])));
  end;

{$endif}


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  vocalTract
  //
  //  purpose:  Updates the pressure wave throughout the vocal tract,
  //                        and returns the summed output of the oral and nasal
  //                        cavities.  Also injects frication appropriately.
  //
  //        arguments:      input, frication
  //
  //  internal
  //  functions:  reflectionFilter, radiationFilter,
  //                        nasalReflectionFilter, nasalRadiationFilter
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function vocalTract( input: Double; frication: Double): Double;
  var
    i, j, k : Integer;
    delta, junctionPressure: Double;
  begin
    // INCREMENT CURRENT AND PREVIOUS POINTERS

    Inc( current_ptr);
    Inc( prev_ptr   );

    if current_ptr > 1
    then current_ptr := 0;

    if prev_ptr > 1
    then prev_ptr := 0;

    // UPDATE OROPHARYNX
    // INPUT TO TOP OF TUBE
    oropharynx[ S1][ TOP][ current_ptr] := ( oropharynx[ S1][ BOTTOM][ prev_ptr] * dampingFactor) + input;

    // CALCULATE THE SCATTERING JUNCTIONS FOR S1-S2
    delta := oropharynx_coeff[ C1] * ( oropharynx[ S1][ TOP][ prev_ptr] - oropharynx[ S2][ BOTTOM][ prev_ptr]);
    oropharynx[ S2][ TOP][ current_ptr] := ( oropharynx[ S1][ TOP][ prev_ptr] + delta) * dampingFactor;
    oropharynx[ S1][ BOTTOM][ current_ptr] := ( oropharynx[ S2][ BOTTOM][ prev_ptr] + delta) * dampingFactor;

    // CALCULATE THE SCATTERING JUNCTIONS FOR S2-S3 AND S3-S4

    j := C2;
    k := FC1;

    for i := S2 to S4 - 1
    do begin
      delta := oropharynx_coeff[ j] * ( oropharynx[ i][ TOP][ prev_ptr] - oropharynx[ i+1][ BOTTOM][ prev_ptr]);
      oropharynx[ i+1][ TOP][ current_ptr] := (( oropharynx[ i][ TOP][ prev_ptr] + delta) * dampingFactor) + ( fricationTap[ k] * frication);
      oropharynx[ i][ BOTTOM][ current_ptr] := ( oropharynx[ i+1][ BOTTOM][ prev_ptr] + delta) * dampingFactor;
      Inc( j);
      Inc( k);
     end;

    // UPDATE 3-WAY JUNCTION BETWEEN THE MIDDLE OF R4 AND NASAL CAVITY
    junctionPressure := ( alpha[ LEFT] * oropharynx[ S4][ TOP][ prev_ptr])+ ( alpha[ RIGHT] * oropharynx[ S5][ BOTTOM][ prev_ptr]) + ( alpha[ UPPER] * nasal[ VELUM][ BOTTOM][ prev_ptr]);
    oropharynx[ S4][ BOTTOM][ current_ptr] := ( junctionPressure - oropharynx[ S4][ TOP][ prev_ptr]) * dampingFactor;
    oropharynx[ S5][ TOP][ current_ptr] := (( junctionPressure - oropharynx[ S5][ BOTTOM][ prev_ptr]) * dampingFactor) + ( fricationTap[ FC3] * frication);
    nasal[ VELUM][ TOP][ current_ptr] := ( junctionPressure - nasal[ VELUM][ BOTTOM][ prev_ptr]) * dampingFactor;

    // CALCULATE JUNCTION BETWEEN R4 AND R5 ( S5-S6)
    delta := oropharynx_coeff[ C4] * ( oropharynx[ S5][ TOP][ prev_ptr] - oropharynx[ S6][ BOTTOM][ prev_ptr]);
    oropharynx[ S6][ TOP][ current_ptr] := (( oropharynx[ S5][ TOP][ prev_ptr] + delta) * dampingFactor) + ( fricationTap[ FC4] * frication);
    oropharynx[ S5][ BOTTOM][ current_ptr] := ( oropharynx[ S6][ BOTTOM][ prev_ptr] + delta) * dampingFactor;

    // CALCULATE JUNCTION INSIDE R5 ( S6-S7) ( PURE DELAY WITH DAMPING)
    oropharynx[ S7][ TOP][ current_ptr] := ( oropharynx[ S6][ TOP][ prev_ptr] * dampingFactor) + ( fricationTap[ FC5] * frication);
    oropharynx[ S6][ BOTTOM][ current_ptr] := oropharynx[ S7][ BOTTOM][ prev_ptr] * dampingFactor;

    // CALCULATE LAST 3 INTERNAL JUNCTIONS ( S7-S8, S8-S9, S9-S10)

    j := C5;
    k := FC6;

    for i := S7 to S10 - 1
    do begin
      delta := oropharynx_coeff[ j] * ( oropharynx[ i][ TOP][ prev_ptr] - oropharynx[ i+1][ BOTTOM][ prev_ptr]);
      oropharynx[ i+1][ TOP][ current_ptr] := (( oropharynx[ i][ TOP][ prev_ptr] + delta) * dampingFactor) + ( fricationTap[ k] * frication);
      oropharynx[ i][ BOTTOM][ current_ptr] := ( oropharynx[ i+1][ BOTTOM][ prev_ptr] + delta) * dampingFactor;
      Inc( j);
      Inc( k);
    end;

    // REFLECTED SIGNAL AT MOUTH GOES THROUGH A LOWPASS FILTER
    oropharynx[ S10][ BOTTOM][ current_ptr] :=  dampingFactor * reflectionFilter( oropharynx_coeff[ C8] * oropharynx[ S10][ TOP][ prev_ptr]);

    // OUTPUT FROM MOUTH GOES THROUGH A HIGHPASS FILTER
    Result := radiationFilter(( 1.0 + oropharynx_coeff[ C8]) * oropharynx[ S10][ TOP][ prev_ptr]);


    // UPDATE NASAL CAVITY

    j := NC1;

    for i := VELUM to N6 - 1
    do begin
      delta := nasal_coeff[ j] * ( nasal[ i][ TOP][ prev_ptr] - nasal[ i+1][ BOTTOM][ prev_ptr]);
      nasal[ i+1][ TOP][ current_ptr] := ( nasal[ i][ TOP][ prev_ptr] + delta) * dampingFactor;
      nasal[ i][ BOTTOM][ current_ptr] := ( nasal[ i+1][ BOTTOM][ prev_ptr] + delta) * dampingFactor;
      Inc( j);
    end;

    // REFLECTED SIGNAL AT NOSE GOES THROUGH A LOWPASS FILTER
    nasal[ N6][ BOTTOM][ current_ptr] := dampingFactor * nasalReflectionFilter( nasal_coeff[ NC6] * nasal[ N6][ TOP][ prev_ptr]);

    // OUTPUT FROM NOSE GOES THROUGH A HIGHPASS FILTER
    Result := Result + nasalRadiationFilter(( 1.0 + nasal_coeff[ NC6]) * nasal[ N6][ TOP][ prev_ptr]);

    // RETURN SUMMED OUTPUT FROM MOUTH AND NOSE   // == Result
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  throat
  //
  //  purpose:  Simulates the radiation of sound through the walls
  //                        of the throat.  Note that this form of the filter
  //                        uses addition instead of subtraction for the
  //                        second term, since tb1 has reversed sign.
  //
  //        arguments:      input
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function throat( input: Double): Double;
  begin
    Result  := ( ta0 * input) + ( tb1 * throatY);
    throatY := Result;
    Result  := Result * throatGain;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  bandpassFilter
  //
  //  purpose:  Frication bandpass filter, with variable center
  //                        frequency and bandwidth.
  //
  //        arguments:      input
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function bandpassFilter( input: Double): Double;
  begin
    Result := 2.0 * (( bpAlpha * ( input - xn2)) + ( bpGamma * yn1) - ( bpBeta * yn2));
    xn2    := xn1;
    xn1    := input;
    yn2    := yn1;
    yn1    := Result;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //        aFunction:       amplitude
  //
  //        purpose:        Converts dB value to amplitude value.
  //
  //        internal
  //        functions:      none
  //
  //        aLibrary
  //        functions:      pow
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function amplitude( decibelLevel: Double): Double;
  begin
    // printf( "Value passed to amplitude tube.c:1647 is %f", decibelLevel);
    // CONVERT 0-60 RANGE TO -60-0 RANGE
    decibelLevel := decibelLevel - VOL_MAX;

    // IF -60 OR LESS, RETURN AMPLITUDE OF 0
    if decibelLevel <= -VOL_MAX
    then Result := 0.0

    // IF 0 OR GREATER, RETURN AMPLITUDE OF 1
    else if decibelLevel >= 0.0
    then Result := 1.0

    // ELSE RETURN INVERSE LOG VALUE
    else Result := ( Power( 10.0, ( decibelLevel / 20.0)));
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //        aFunction:       frequency
  //
  //        purpose:        Converts a given pitch ( 0 = middle C) to the
  //                        corresponding frequency.
  //
  //        internal
  //        functions:      none
  //
  //        aLibrary
  //        functions:      pow
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function frequency( pitch: Double): Double;
  begin
    Result := PITCH_BASE * Power( 2.0, (( pitch + PITCH_OFFSET) / 12.0));
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  maximallyFlat
  //
  //  purpose:  Calculates coefficients for a linear phase lowpass FIR
  //                        filter, with beta being the center frequency of the
  //                        transition band ( as a fraction of the sampling
  //                        frequency), and gamme the width of the transition
  //                        band.
  //
  //        arguments:      beta, gamma, np, coefficient
  //
  //  internal
  //  functions:  rationalApproximation
  //
  //  aLibrary
  //  functions:  cos, pow
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function maximallyFlat( beta: Double; gamma: Double; var np: Integer; coefficient : PDoubles): Integer;
  var
    a : TCoefficients;
    c : TCoefficients;
    betaMinimum, ac : Double;
    nt, numerator, n, ll, i : Integer;
    j, jj, m : Integer;
    x, sum, y, z: Double;
  begin
    // INITIALIZE NUMBER OF POINTS
    np := 0;

    // CUT-OFF FREQUENCY MUST BE BETWEEN 0 HZ AND NYQUIST
    if   ( beta <= 0.0) or ( beta >= 0.5)
    then begin
      Result := BETA_OUT_OF_RANGE;
      Exit;
    end;

    // TRANSITION BAND MUST FIT WITH THE STOP BAND
    // betaMinimum := (( 2.0 * beta) < ( 1.0 - 2.0 * beta)) ? ( 2.0 * beta) : ( 1.0 - 2.0 * beta);

    if   ( 2.0 * beta) < ( 1.0 - 2.0 * beta)
    then betaMinimum := 2.0 * beta
    else betaMinimum := 1.0 - 2.0 * beta;

    if   ( gamma <= 0.0) or ( gamma >= betaMinimum)
    then begin
      Result := GAMMA_OUT_OF_RANGE;
      Exit;
    end;

    // MAKE SURE TRANSITION BAND NOT TOO SMALL
    nt := Floor( 1.0 / ( 4.0 * gamma * gamma));

    if   nt > 160
    then begin
      Result := GAMMA_TOO_SMALL;
      Exit;
    end;

    // CALCULATE THE RATIONAL APPROXIMATION TO THE CUT-OFF POINT
    ac := ( 1.0 + cos( 2 * PI * beta)) / 2.0;
    rationalApproximation( ac, nt, numerator, np);

    // CALCULATE FILTER ORDER
    n := ( 2 * ( np)) - 1;

    if   numerator = 0
    then numerator := 1;


    // COMPUTE MAGNITUDE AT NP POINTS
    c[ 1] := 1.0;
    a[ 1] := 1.0;
    ll   := nt - numerator;

    for i := 2 to np       // note .. to np, not np - 1
    do begin
      sum := 1.0;
      c[ i] := cos( 2 * PI * (( i-1)/n));
      x := ( 1.0 - c[ i]) / 2.0;
      y := x;

      if   numerator = nt
      then continue;

      for j := 1 to ll      // note .. to ll, not ll - 1
      do begin
        z := y;

        if numerator <> 1
        then begin
          for jj := 1 to numerator - 1
          do z := z * 1.0 + ( j / jj);
        end;
        y   := y * x;
        sum := z + z;
      end;

      a[ i] := sum * Power(( 1.0 - x), numerator);
    end;


    // CALCULATE WEIGHTING COEFFICIENTS BY AN N-POINT IDFT
    for i := 1 to np
    do begin
      coefficient[ i] := a[ 1] / 2.0;

      for j := 2 to np
      do begin
        m := (( i - 1) * ( j - 1)) mod n;

        if   m > nt
        then m := n - m;

        coefficient[ i] := coefficient[ i] + c[ m+1] * a[ j];
      end;

      coefficient[ i]:= coefficient[ i] * 2.0/n;
    end;

    Result := 0;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  trim
  //
  //  purpose:  Trims the higher order coefficients of the FIR filter
  //                        which fall below the cutoff value.
  //
  //        arguments:      cutoff, numberCoefficients, coefficient
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  fabs
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure trim( cutoff: Double; var numberCoefficients: Integer; coefficient: PDoubles);
  begin
    while numberCoefficients > 0
    do begin
      if abs( coefficient^[ numberCoefficients]) >= abs( cutoff)
      then Exit
      else numberCoefficients := numberCoefficients - 1;
    end;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  rationalApproximation
  //
  //  purpose:  Calculates the best rational approximation to 'number',
  //                        given the maximum 'order'.
  //
  //        arguments:      number, order, numerator, denominator
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  fabs
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure rationalApproximation( number: Double; var order, numerator, denominator: Integer);
  var
    fractionalPart, minimumError: Double;
    i, orderMaximum, modulus: Integer;
    ps : Double;
    ip : Integer;
    error : Double;
  begin
    minimumError := 1.0;
    modulus      := 0;

    // RETURN IMMEDIATELY IF THE ORDER IS LESS THAN ONE

    if order <= 0
    then begin
      numerator   := 0;
      denominator := 0;
      order       := -1;
      Exit;
    end;

    // FIND THE ABSOLUTE VALUE OF THE FRACTIONAL PART OF THE NUMBER
    fractionalpart := abs( number - Floor( number));

    // DETERMINE THE MAXIMUM VALUE OF THE DENOMINATOR
    orderMaximum := 2 * order;

    // orderMaximum := ( orderMaximum > LIMIT) ? LIMIT : orderMaximum;
    if orderMaximum > LIMIT
    then orderMaximum := LIMIT;

    // FIND THE BEST DENOMINATOR VALUE
    for i := order to orderMaximum
    do begin
      ps    := i * fractionalPart;
      ip    := Floor( ps + 0.5);
      error := abs(( ps - ip) / i);

      if error < minimumError
      then
      begin
        minimumError := error;
        modulus      := ip;
        denominator  := i;
      end;
    end;

    // DETERMINE THE NUMERATOR VALUE, MAKING IT NEGATIVE IF NECESSARY
    numerator := Floor( Abs( number)) * denominator + modulus;

    if number < 0
    then numerator := - numerator;

    // SET THE ORDER
    order := denominator - 1;

    // RESET THE NUMERATOR AND DENOMINATOR IF THEY ARE EQUAL
    if numerator = denominator
    then begin
      denominator := orderMaximum;
      numerator   := denominator - 1;
      order       := numerator;
    end;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  FIRFilter
  //
  //  purpose:  Is the linear phase, lowpass FIR filter.
  //
  //        arguments:      input, needOutput
  //
  //  internal
  //  functions:  increment, decrement
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function FIRFilter( input: Double; needOutput: Integer): Double;
  var
    i : Integer;
  begin
    if needOutput <> 0
    then begin
      Result := 0.0;

      // PUT INPUT SAMPLE INTO DATA BUFFER
      FIRData^[ FIRPtr] := input;

      // SUM THE OUTPUT FROM ALL FILTER TAPS
      for i := 0 to numberTaps - 1
      do begin
        Result := Result + FIRData^[ FIRPtr] * FIRCoef^[ i];
        FIRPtr := increment( FIRPtr, numberTaps);
      end;

      // DECREMENT THE DATA POINTER READY FOR NEXT CALL
      FIRPtr := decrement( FIRPtr, numberTaps);

      // RETURN THE OUTPUT VALUE    // Implicit .. in Result
    end
    else begin
      // PUT INPUT SAMPLE INTO DATA BUFFER
      FIRData^[ FIRPtr] := input;

      // ADJUST THE DATA POINTER, READY FOR NEXT CALL
      FIRPtr := decrement( FIRPtr, numberTaps);

      Result := 0.0;
    end;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  increment
  //
  //  purpose:  Increments the aPointer to the circular FIR filter
  //                        buffer, keeping it in the range 0 ^. modulus-1.
  //
  //        arguments:      aPointer, modulus
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function increment( aPointer: Integer; modulus: Integer): Integer;
  begin
    if   aPointer + 1 >= modulus
    then Result := 0
    else Result := aPointer + 1;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  decrement
  //
  //  purpose:  Decrements the aPointer to the circular FIR filter
  //                        buffer, keeping it in the range 0 ^. modulus-1.
  //
  //        arguments:      aPointer, modulus
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function decrement( aPointer: Integer; modulus: Integer): Integer;
  begin
    if   aPointer - 1 < 0
    then Result := modulus-1
    else Result := aPointer - 1;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  initializeConversion
  //
  //  purpose:  Initializes all the sample rate conversion functions.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  initializeFilter, initializeBuffer
  //
  //  aLibrary
  //  functions:  rint, pow
  //  [ apparently not called 2009-01-07]
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure initializeConversion;
  var
    roundedSampleRateRatio: Double;
  begin
    // INITIALIZE FILTER IMPULSE RESPONSE;
    initializeFilter;

    // CALCULATE SAMPLE RATE RATIO
    sampleRateRatio := outputRate / sampleRate;
    // printf( 'tube.c:2047 output-rate is: %f, sample rate is: %d, sample rate ratio is: %f\n',outputRate, sampleRate, sampleRateRatio);

    // CALCULATE TIME REGISTER INCREMENT
    timeRegisterIncrement := Round( Power( 2.0, FRACTION_BITS) / sampleRateRatio);

    // CALCULATE ROUNDED SAMPLE RATE RATIO
    roundedSampleRateRatio := Power( 2.0, FRACTION_BITS) / timeRegisterIncrement;

    // CALCULATE PHASE OR FILTER INCREMENT
    if   sampleRateRatio >= 1.0
    then filterIncrement := L_RANGE
    else phaseIncrement  := Round( sampleRateRatio * FRACTION_RANGE);

    // CALCULATE PAD SIZE
    // padSize := ( sampleRateRatio >= 1.0) ? ZERO_CROSSINGS : ( Integer)(( Double)ZERO_CROSSINGS / roundedSampleRateRatio) + 1;

    if sampleRateRatio >= 1.0
    then padSize := ZERO_CROSSINGS
    else padSize := Floor(( ZERO_CROSSINGS / roundedSampleRateRatio) + 1);

    // INITIALIZE THE RING BUFFER
    initializeBuffer;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  initializeFilter
  //
  //  purpose:  Initializes filter impulse response and impulse delta
  //                        values.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  sin, cos
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure initializeFilter;
  var
    x, y, IBeta : Double;
    i           : Integer;
    temp        : Double;
  begin
    // INITIALIZE THE FILTER IMPULSE RESPONSE
    h[ 0] := LP_CUTOFF;
    x     := PI / L_RANGE;

    for i := 1 to FILTER_LENGTH - 1
    do begin
      y     := i * x;
      h[ i] := Sin( y * LP_CUTOFF) / y;
    end;

    // APPLY A KAISER WINDOW TO THE IMPULSE RESPONSE;
    IBeta := 1.0 / Izero2( BETA);

    for i := 0 to FILTER_LENGTH - 1
    do begin
      temp  := i / FILTER_LENGTH;
      h[ i] := h[ i] * Izero2( BETA * sqrt( 1.0 - ( temp * temp))) * IBeta;
    end;

    // INITIALIZE THE FILTER IMPULSE RESPONSE DELTA VALUES
    for i := 0 to FILTER_LIMIT - 1
    do  deltaH[ i] := h[ i+1] - h[ i];

    deltaH[ FILTER_LIMIT] := 0.0 - h[ FILTER_LIMIT];
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  Izero2
  //
  //  purpose:  Returns the value for the modified Bessel aFunction of
  //                        the first kind, order 0, as a Double.
  //
  //        arguments:      x - input argument
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  function Izero2( x: Double): Double;
  var
    sum, u, halfx, temp : Double;
    n : Integer;
  begin
    sum   := 1;
    u     := 1;
    n     := 1;
    halfx := x / 2.0;

    repeat
      temp := halfx / n;
      n    := n + 1;
      temp := temp * temp;
      u    := u * temp;
      sum  := sum + u;
    until u < IzeroEPSILON * sum;

    Result := ( sum);
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  initializeBuffer
  //
  //  purpose:  Initializes the ring buffer used for sample rate
  //                        conversion.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure initializeBuffer;
  var
    i : Integer;
  begin
    // FILL THE RING BUFFER WITH ALL ZEROS
    for i := 0 to BUFFER_SIZE - 1
    do buffer[ i] := 0.0;

    // INITIALIZE FILL POINTER
    fillPtr := padSize;

    // CALCULATE FILL SIZE
    fillSize := BUFFER_SIZE - ( 2 * padSize);
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  dataFill
  //
  //  purpose:  Fills the ring buffer with one sample, increments
  //                        the counters and pointers, and empties the buffer when
  //                        full.
  //
  //        arguments:      data
  //
  //  internal
  //  functions:  srIncrement, dataEmpty
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure dataFill( data: Double);
  begin
    // PUT THE DATA INTO THE RING BUFFER;
    buffer[ fillPtr] := data;

    // INCREMENT THE FILL POINTER: ( ;  MODULO THE BUFFER SIZE
    srIncrement( fillPtr, BUFFER_SIZE);

    // INCREMENT THE COUNTER, AND EMPTY THE BUFFER IF FULL

    Inc( fillCounter);

    if   fillCounter >= fillSize
    then begin
      dataEmpty;
      // RESET THE FILL COUNTER
      fillCounter := 0;
    end;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  dataEmpty
  //
  //  purpose:  Converts available portion of the input signal to the
  //                        new sampling rate, and outputs the samples to the
  //                        sound .
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  srDecrement, srIncrement
  //
  //  aLibrary
  //  functions:  Round, Abs,
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure dataEmpty;
  var
    endPtr : Integer;
    index : Integer;
    phaseIndex, impulseIndex : Cardinal;
    absoluteSampleValue, output, impulse : Double;
    filterIndex: Cardinal;
    interpolation : Double;
  begin
    // CALCULATE END POINTER
    endPtr := fillPtr - padSize;

    // ADJUST THE END POINTER: IF LESS THAN ZERO
    if   endPtr < 0
    then endPtr := endPtr + BUFFER_SIZE;

    // ADJUST THE ENDPOINT, IF LESS THAN THE EMPTY POINTER
    if   endPtr < emptyPtr
    then endPtr := endPtr + BUFFER_SIZE;


    // UPSAMPLE LOOP ( SLIGHTLY MORE EFFICIENT THAN DOWNSAMPLING)
    if   sampleRateRatio >= 1.0
    then begin
      while emptyPtr < endPtr
      do begin

        // RESET ACCUMULATOR TO ZERO;
        output := 0.0;

        // CALCULATE INTERPOLATION VALUE ( STATIC WHEN UPSAMPLING
        interpolation := mValue( timeRegister) / M_RANGE;

        // COMPUTE THE LEFT SIDE OF THE FILTER CONVOLUTION;
        index       := emptyPtr;
        filterIndex := lValue( timeRegister);

        while filterIndex < FILTER_LENGTH
        do begin
          output := output + ( buffer[ index] * ( h[ filterIndex] + ( deltaH[ filterIndex] * interpolation)));
          srDecrement( index, BUFFER_SIZE);
          filterIndex := filterIndex + filterIncrement;
        end;

        // ADJUST VALUES FOR RIGHT SIDE CALCULATION
        timeRegister  := not timeRegister;
        interpolation := mValue( timeRegister) / M_RANGE;

        // COMPUTE THE RIGHT SIDE OF THE FILTER CONVOLUTION
        index := emptyPtr;
        srIncrement( index,BUFFER_SIZE);
        filterIndex := lValue( timeRegister);

        while filterIndex < FILTER_LENGTH
        do begin
          output      := output + ( buffer[ index] * ( h[ filterIndex] + ( deltaH[ filterIndex] * interpolation)));
          filterIndex := filterIndex + filterIncrement;
        end;


        // RECORD MAXIMUM SAMPLE VALUE
        absoluteSampleValue := Abs( output);

        if   absoluteSampleValue > maximumSampleValue
        then maximumSampleValue := absoluteSampleValue;

        // INCREMENT SAMPLE NUMBER
        Inc( numberSamples);
        putCircBuff2( output);     // if room, put output in circBuff2

        // CHANGE TIME REGISTER BACK TO ORIGINAL FORM
        timeRegister := not timeRegister;

        // INCREMENT THE TIME REGISTER
        timeRegister := timeRegister + timeRegisterIncrement;

        // INCREMENT THE EMPTY POINTER, ADJUSTING IT AND END POINTER
        emptyPtr := emptyPtr + Integer( nValue( timeRegister));

        if emptyPtr >= BUFFER_SIZE
        then begin
          emptyPtr := emptyPtr - BUFFER_SIZE;
          endPtr   := endPtr   - BUFFER_SIZE;
        end;

        // CLEAR N PART OF TIME REGISTER
        timeRegister := timeRegister and ( not N_MASK);
      end;
      // DOWNSAMPLING CONVERSION LOOP
    end
    else begin
      while emptyPtr < endPtr
      do begin

        // RESET ACCUMULATOR TO ZERO
        output := 0.0;

        // COMPUTE P PRIME
        phaseIndex := Round(( fractionValue( timeRegister)) * sampleRateRatio);

        // COMPUTE THE LEFT SIDE OF THE FILTER CONVOLUTION
        index := emptyPtr;

        impulseIndex := phaseIndex shr M_BITS;

        while impulseIndex < FILTER_LENGTH
        do begin
          impulse := h[ impulseIndex] + ( deltaH[ impulseIndex] * (( mValue( phaseIndex)) / M_RANGE));
          output  := output + ( buffer[ index] * impulse);
          srDecrement( index, BUFFER_SIZE);
          phaseIndex   := phaseIndex + phaseIncrement;
          impulseIndex := phaseIndex shr M_BITS;
        end;

        // COMPUTE P PRIME, ADJUSTED FOR RIGHT SIDE
        phaseIndex := Round( fractionValue( not timeRegister) * sampleRateRatio);

        // COMPUTE THE RIGHT SIDE OF THE FILTER CONVOLUTION
        index := emptyPtr;
        srIncrement( index, BUFFER_SIZE);

        impulseIndex := phaseIndex shr M_BITS;

        while impulseIndex < FILTER_LENGTH
        do begin
          impulse := h[ impulseIndex] + ( deltaH[ impulseIndex] * ( mValue( phaseIndex) / M_RANGE));
          output  := output + ( buffer[ index] * impulse);
          srIncrement( index, BUFFER_SIZE);
          phaseIndex := phaseIndex + phaseIncrement;
          impulseIndex := phaseIndex shr M_BITS;
        end;

        // RECORD MAXIMUM SAMPLE VALUE
        absoluteSampleValue := Abs( output);

        if   absoluteSampleValue > maximumSampleValue
        then maximumSampleValue := absoluteSampleValue;

        // INCREMENT SAMPLE NUMBER
        numberSamples:= numberSamples + 1;
        putCircBuff2( output);     // if room, put output in circBuff2

        // INCREMENT THE TIME REGISTER
        timeRegister := timeRegister + timeRegisterIncrement;

        // INCREMENT THE EMPTY POINTER, ADJUSTING IT AND END POINTER
        emptyPtr := emptyPtr + Integer( nValue( timeRegister));

        if   emptyPtr >= BUFFER_SIZE
        then begin
          emptyPtr := emptyPtr - BUFFER_SIZE;
          endPtr   := endPtr   - BUFFER_SIZE;
        end;

        // CLEAR N PART OF TIME REGISTER
        timeRegister := timeRegister and ( not N_MASK);
      end;
    end;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  flushBuffer
  //
  //  purpose:  Pads the buffer with zero samples, and flushes it by
  //                        converting the remaining samples.
  //
  //        arguments:      none
  //
  //  internal
  //  functions:  dataFill, dataEmpty
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure flushBuffer;
  var
    i : Integer;
  begin
    // PAD END OF RING BUFFER WITH ZEROS;
    for i := 0 to 2 * padSize - 1
    do dataFill( 0.0);

    // FLUSH UP TO FILL POINTER - PADSIZE
    dataEmpty;
  end;


  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //  aFunction:  srIncrement
  //
  //  purpose:  Increments the aPointer, keeping it within the range
  //                        0 to ( modulus-1).
  //
  //        arguments:      aPointer, modulus
  //
  //  internal
  //  functions:  none
  //
  //  aLibrary
  //  functions:  none
  //
  // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure srIncrement( var aPointer: Integer; modulus: Integer);
  begin
    aPointer := aPointer + 1;

    if   aPointer >= modulus
    then aPointer := aPointer - modulus;
  end;


  procedure srDecrement( var aPointer: Integer; modulus: Integer);
  begin
    aPointer := aPointer - 1;

    if   aPointer < modulus
    then aPointer := aPointer + modulus;
  end;


  procedure ConnectLogger( const aLogger: TTubeLogger);
  begin
    GLogger := aLogger;
  end;


  // FUNCTIONS TO ALLOW OBJECTIVE-C TO SET TUBE PARAMETERS;

  procedure setGlotPitch( value: Double);
  begin
    current.glotPitch := value;
  end;


  procedure setGlotVol( value: Double);
  begin
    current.glotVol := value;
  end;


  procedure setAspVol( value: Double);
  begin
    current.aspVol := value;
  end;


  procedure setFricVol( value: Double);
  begin
    current.fricVol := value;
  end;


  procedure setfricPos( value: Double);
  begin
    current.fricPos := value;
  end;


  procedure setFricCF( value: Double);
  begin;
    current.fricPos := value;
  end;


  procedure setFricBW( value: Double);
  begin
    current.fricBW := value;
  end;


  procedure setRadius( value: Double; index: Integer);
  begin
    current.radius[ index] := value;
  end;


  procedure setVelum( value: Double);
  begin
    current.velum := value;
  end;


  procedure setVolume( value: Double);
  begin
    volume := value;
  end;


  procedure setWaveformType( value: Integer);
  begin
    waveform := value;
  end;


  procedure setTp( value: Double);
  begin
    tp := value;;
  end;


  procedure setTnMin( value: Double);
  begin
    tnMin := value;
  end;


  procedure setTnMax( value: Double);
  begin
    tnMax := value;
  end;


  procedure setBreathiness( value: Double);
  begin
    breathiness := value;
  end;


  procedure setTubeLength( value: Double);
  begin
    tubeLength := value;
  end;


  procedure setTemperature( value: Double);
  begin
    temperature := value;
  end;


  procedure setLossFactor( value: Double);
  begin
    lossFactor := value;
  end;


  procedure setApScale( value: Double);
  begin
    apScale := value;
  end;


  procedure setMouthCoef( value: Double);
  begin
    mouthCoef := value;
  end;


  procedure setNoseCoef( value: Double);
  begin
    noseCoef := value;
  end;


  procedure setNoseRadius( value: Double; index: Integer);
  begin
    noseRadius[ index] := value;
  end;


  procedure setThroatCutoff( value: Double);
  begin
    throatCutoff := value;
  end;


  procedure setModulation( value: Integer);
  begin;
    modulation := value;
  end;


  procedure setMixOffset( value: Double);
  begin
    mixOffset := value;
  end;


  // FUNCTIONS TO ALLOW INTERFACE OBJECTIVE-C ACCESS TO DEFAULT TUBE PARAMETERS;

  function  getGlotPitchDefault: Double;
  begin;
    Result := originalDefaults.glotPitch;
  end;


  function  getGlotVolDefault: Double;
  begin
    Result := originalDefaults.glotVol;
  end;


  function  getAspVolDefault: Double;
  begin
    Result := originalDefaults.aspVol;
  end;


  function  getFricVolDefault: Double;
  begin
    Result := originalDefaults.fricVol;
  end;


  function  getFricPosDefault: Double;
  begin
    Result := originalDefaults.fricPos;
  end;


  function  getFricCFDefault: Double;
  begin
    Result := originalDefaults.fricCF;
  end;


  function  getFricBWDefault: Double;
  begin
    Result := originalDefaults.fricBW;
  end;


  function  getRadiusDefault( index: Integer): Double;
  begin
    Result := originalDefaults.radius[ index];
  end;


  function  getVelumRadiusDefault: Double;
  begin
    Result := originalDefaults.velum;
  end;


  function  getVolumeDefault: Double;
  begin
    Result := volume;
  end;


  function  getWaveformDefault: Integer;
  begin
    Result := waveform;
  end;


  function  getBalanceDefault: Double;
  begin
    Result := balance;
  end;


  function  getTpDefault: Double;
  begin
    Result := tp;
  end;


  function  getTnMinDefault: Double;
  begin
    Result := tnMin;
  end;


  function  getTnMaxDefault: Double;
  begin
    Result := tnMax;
  end;


  function  getBreathinessDefault: Double;
  begin
    Result := breathiness;
  end;


  function  getTubeLengthDefault: Double;
  begin
    Result := tubeLength;
  end;


  function  getTemperatureDefault: Double;
  begin
    Result := temperature;
  end;


  function  getLossFactorDefault: Double;
  begin
    Result := lossFactor;
  end;


  function  getApScaleDefault: Double;
  begin
    Result := apScale;
  end;


  function  getMouthCoefDefault: Double;
  begin
    Result := mouthCoef;
  end;


  function  getNoseCoefDefault: Double;
  begin
    Result := noseCoef;
  end;


  function  getNoseRadiusDefault( index: Integer): Double;
  begin
    Result := noseRadiusOriginalDefaults[ index];
  end;


  function  getThroatCutoffDefault: Double;
  begin
    Result := throatCutoff;
  end;


  function  getThroatVolDefault: Double;
  begin
    Result := throatVol;
  end;


  function  getModulationDefault: Integer;
  begin
    Result := modulation;
  end;


  function  getMixOffsetDefault: Double;
  begin
    Result := mixOffset;
  end;

  // FUNCTIONS TO ALLOW INTERFACE OBJECTIVE-C ACCESS TO TUBE PARAMETERS

  function  getGlotPitch: Double;
  begin
    Result := current.glotPitch;
  end;


  function  getGlotVol: Double;
  begin
    Result := current.glotVol;
  end;


  function  getAspVol: Double;
  begin
    Result := current.aspVol;
  end;


  function  getFricVol: Double;
  begin
    Result := current.fricVol;
  end;


  function  getFricPos: Double;
  begin
    Result := current.fricPos;
  end;


  function  getFricCF: Double;
  begin
    Result := current.fricCF;
  end;


  function  getFricBW: Double;
  begin
    Result := current.fricBW;
  end;


  function  getRadius( index: Integer): Double;
  begin
    Result := current.radius[ index];
  end;


  function  getVelumRadius: Double;
  begin
    Result := current.velum;
  end;


  function  getVolume: Double;
  begin
    Result := volume;
  end;


  function  getWaveform: Integer;
  begin
    Result := waveform;
  end;


  function  getBalance: Double;
  begin
    Result := balance;
  end;


  function  getTp: Double;
  begin
    Result := tp;
  end;


  function  getTnMin: Double;
  begin
    Result := tnMin;
  end;


  function  getTnMax: Double;
  begin
    Result := tnMax;
  end;


  function  getBreathiness: Double;
  begin
    Result := breathiness;
  end;


  function  getTubeLength: Double;
  begin
    Result := tubeLength;
  end;


  function  getTemperature: Double;
  begin
    Result := temperature;
  end;


  function  getLossFactor: Double;
  begin
    Result := lossFactor;
  end;


  function  getApScale: Double;
  begin
    Result := apScale;
  end;


  function  getMouthCoef: Double;
  begin
    Result := mouthCoef;
  end;


  function  getNoseCoef: Double;
  begin
    Result := noseCoef;
  end;


  function  getNoseRadius( index: Integer): Double;
  begin
    Result := noseRadius[ index];
  end;


  function  getThroatCutoff: Double;
  begin
    Result := throatCutoff;
  end;


  function  getThroatVol: Double;
  begin
    Result := throatVol;
  end;


  function  getModulation: Integer;
  begin
    Result := modulation;
  end;


  function  getMixOffset: Double;
  begin
    Result := mixOffset;
  end;


  function  getActualTubeLength: Double;
  begin
    Result := actualTubeLength;
  end;


  function  getSampleRate: Integer;
  begin
    // printf( "Sample rate in get routine is %f\n", sampleRate);
    Result := sampleRate;
  end;


  function  getControlPeriod: Integer;
  begin
    // printf( "Control period in get routine is %f\n", controlPeriod);
    Result := controlPeriod;
  end;


  function  getControlRate: Double;
  begin
    Result := controlRate;
  end;


  function  getWavetable( index: Integer): Double;
  begin
    Result := wavetable[ index];
  end;


  procedure initCircBuff;
  begin
    CircularBuffer := TCircularBuffer.Create( CIRC_BUFF_SIZE);
  end;


  procedure putCircBuff( value : Double);
  begin
    CircularBuffer.Push( value);
  end;

  function getCircBuff: Double;
  begin
    Result := CircularBuffer.Pop;
  end;


  procedure initCircBuff2;
  begin
    CircularBuffer2 := TCircularBuffer.Create( CIRC_BUFF2_SIZE);
  end;


  procedure putCircBuff2( value: Double);
  begin
    CircularBuffer2.Push( value);
  end;


  function getCircBuff2: Double;
  begin
    Result := CircularBuffer2.Pop;
  end;


initialization

finalization

  FreeAndNil( CircularBuffer );
  FreeAndNil( CircularBuffer2);

end.

